/*
 *   WDX: WitchDoctor-WDX, versatile distributed lock manager
 *   Copyright (C) 2003 Dominique Chabord,
 *   Please refer to credits section below
 *
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ************************************************
 *
 *   History and credits:
 *   First version of WitchDoctor-WDX was based on:
 *
 *   DECwatchdog/Autopilot, watchdog and lock manager
 *   Copyright (C) 2003 HP
 *
 *   Developed by
 *     Computer Special Systems, Annecy France
 *     A Division of Digital Equipment Corporation
 *   and
 *     Custom-Systems, Annecy France
 *     An engineering center of Compaq
 *
 *   Provided under GNU-GPL terms by HP, 30 June 2003
 *   GNU-GPL replaces previous license statements
 *
 *   Rebranded, ported and modified for Shaman-X project, 15 July 2003
 *   by Dominique Chabord
 *   and maintained by other contributors as mentionned in the code
 *
 *   ************************************************
 *
 *   WDX is promoted by Shaman-X project
 *
 *   Check for new versions and documentation at
 *   http://www.shaman-x.org
 *
 *   ************************************************
 *
 *   Most of WDX code is designed to run unchanged on
 *   different operating systems, including at V1.0 time:
 *   Gnu-Linux, WindowsNT, DigitalUNIX and OpenVMS.
 *   Operating system specifics are identified by dedicated files.
 *
 *   Evolutions
 *   ----------
 *   July-2003     Dominique Chabord    WitchDoctor-WDX V1.0
 *                 Creation
 *
 *   ************************************************
 *
 *
 *	wdx_api.c
 *
 *	Module to provide API to external processes and user interface process
 *
 *      Routines :
 *	int wdx_open()
 *	int wdx_close()
 *	int wdx_read()
 *	int wdx_write()
 *
 ************************************************************************/
/*
 *
 *  15-Feb-94 -- Cornu
 *      Creation
 * MK001 - replaced sleep() by spc_sleep
 * MK002 - Modification required for i386 compilation
 */
/*
 * include files
 */
#undef WDX_API_MODULE
#define WDX_API_MODULE

#include "wdx.h"
#include "wdx_api.h"

#define SYNCHRO_ATTEMPT 5
#define SYNCHRO_WAIT SPECIFIC_SYNCHRO_WAIT
enum wdx_api_used { API_SLOT_UNUSED, API_SLOT_USED };
enum daemon_control { WDX_DAEMON_ALIVE, WDX_DAEMON_DEAD };

struct wdx_application_link
{
   enum   wdx_api_used     use;  /* in use flag */
   char   mission[MISSION_LEN];  /* mission name */
   struct member_base     data;	 /* member data informations */
   enum control_mode       mode; /* how the link is opened */
   int	                socket;  /* socket descriptor */
   int      	          fifo;  /* fifo descriptor */
   int              identifier;  /* current message identifier */
   enum daemon_control control;  /* control flag */
};

/* link table */
static struct wdx_application_link wdxtable[MAX_PROCESS_ON_APPLICATION];
static apiinit=0;

/* Global variable */
int wdxerrno;

/* Local and external routines prototypes */
void reset_application_link(struct wdx_application_link *l);
int find_first_free_slot();
int find_slot_index(char *s);
int find_slot_in_use(int d);
int check_type(int t);
void wdx_api_init();
void spc_sleep();        /* MK002 */


/*
 * LOCAL ROUTINES
 */

void reset_application_link(l)
struct wdx_application_link *l;
{
    l->use        = API_SLOT_UNUSED;
    l->mission[0] = '\0';
    l->socket     = -1;  /* socket descriptor */
    l->fifo       = -1;  /* fifo descriptor */
    l->identifier = 0;  /* fifo descriptor */
    l->control    = WDX_DAEMON_ALIVE; /* by default daemon is alive */
}

/* int find_first_free_slot()
 * Search for the first free index in the wdxtable
 *
 * Argument : None
 *
 * Return   : index if OK
 *	      NOT_AN_INDEX if table full
 */
int find_first_free_slot()
{
int   _i;    /* index in table */
int _ret;    /* returned value */

    _ret = NOT_AN_INDEX;
    for(_i=0;_i<MAX_PROCESS_ON_APPLICATION;_i++)
    {
	if ( wdxtable[_i].use == API_SLOT_UNUSED )
	{
	    _ret = _i;
	    break;
	}
    }
    return(_ret);
}

/* int find_slot_index()
 * Search if a mission is already registered
 *
 * Argument : Pointer to mission name
 *
 * Return   : index if already registered
 *	      NOT_AN_INDEX if not registered
 */
int find_slot_index(s)
char *s;
{
int   _i;    /* index in table */
int _ret;    /* returned value */

    _ret = NOT_AN_INDEX;
    for(_i=0;_i<MAX_PROCESS_ON_APPLICATION;_i++)
    {
	if ( wdxtable[_i].use == API_SLOT_USED )
	{
	     if ( strcmp(wdxtable[_i].mission,s) == 0 )
	     {
	          _ret = _i;
	          break;
	     }
	}
    }
    return(_ret);
}

/* int find_slot_in_use()
 * Search if a descriptor is valid
 *
 * Argument : Pointer to mission name
 *
 * Return   : index if already registered
 *	      NOT_AN_INDEX if not registered
 */
int find_slot_in_use(d)
int d;
{
int _ret;    /* returned value */

    _ret = NOT_AN_INDEX;
    if ( (d <= NOT_AN_INDEX) || ( d >= MAX_PROCESS_ON_APPLICATION ) )
         return(_ret);

    if ( wdxtable[d].use == API_SLOT_USED )
         _ret = d;

    return(_ret);
}

/* int check_type()
 * Check if type is valid
 *
 * Argument : type
 *
 * Return   : 0 OK / -1 error
 */
int check_type(t)
int t;
{
int _ret;    /* returned value */

   _ret = -0;

   switch(t)
   {
   case WDX_GROUP_RUN :         /* run operation */
   case WDX_GROUP_KILL :        /* kill operation */
   case WDX_GROUP_STOP :        /* stop operation */
   case WDX_GROUP_PAUSE :       /* stop operation */
   case WDX_GROUP_START :       /* start operation */
   case WDX_GROUP_EXCEPTION :   /* exception operation */
   case WDX_PROCESS_KILL :      /* leave operation */
   case WDX_MEMBER_UP :         /* join operation */
   case WDX_WAIT_FOR_RESERVE :  /* waitreserve operation */
   case WDX_WAIT_FOR_MOUNT :    /* waitmount operation */
   case WDX_DECLARE :	        /* declare operation */
   case WDX_ERASE :	        /* erase operation */
   case WDX_RESERVE :     	/* reserve operation */
   case WDX_MOUNT :	        /* mount operation */
   case WDX_FREE :              /* free operation */
   case WDX_FAULT :	        /* fault operation */
   case WDX_GET :               /* get ressource operation */
   case WDX_LOG :	        /* log operation */
   case WDX_CONSTANT :          /* version operation : first part */
   case WDX_SCRIPT :            /* version operation : second part */
   case WDX_LIST :	        /* list operation */
   case WDX_MASTERNAME :        /* mastername operation */
   case WDX_STATUS :	        /* status operation */
   case WDX_LOCALNAME :         /* local name operation */
   case WDX_MASTER :	        /* master operation */
   case WDX_NOMASTER :          /* nomaster operation */
   case WDX_GETMASTER :         /* get master operation */
        break;
   default :
        _ret = -1;
   }
   return(_ret);
}

/* void wdx_ast()
 *
 *
 */
void wdx_ast(descriptor)
int descriptor;
{
int slot;

   if ( (slot = find_slot_in_use(descriptor)) != NOT_AN_INDEX )
   {
	wdxtable[slot].control = WDX_DAEMON_DEAD;
	com_cancel_fifo(wdxtable[slot].fifo);
   }

}

/* void wdx_api_init()
 * Initialize API data structures
 *
 * Argument : None
 *
 * Return   : None
 *
 */
void wdx_api_init()
{
int _i;

   for(_i=0;_i<MAX_PROCESS_ON_APPLICATION;_i++)
       reset_application_link(&(wdxtable[_i]));
}

/*
 * PUBLIC ROUTINES : prefixe is "wdx"
 */

/* wdx_open()
 * Open wdx client environment :
 *	- check parameter validity
 *	- check if server information
 *	- get communication link
 *
 * Argument : pt to mission name
 *	      connection mode
 *
 * Return   : slot on success or -1 on error
 *
 * Error    : on error wdxerrno is set to
 *		WDX_ACCES if process is not in superuser group
 *		WDX_INVAL if mission name too long
 *		WDX_INVAL if mission name points to an invalid address
 *		WDX_INVAL if mode is invalid
 *		WDX_EXIST if link already opened
 *		WDX_NOTRUN if mission server not running
 *              WDX_NOTCONN the communication link can't be established
 *		WDX_NOBUFS if communication link can't be created
 */
int wdx_open(missionname,mode)
char *missionname;
int          mode;
{
struct member_base *pt;
struct member_base tmp;
int slot = NOT_AN_INDEX;
int socket;
int fifo;
char fifoname[FIFO_LEN+1];
char daemonlock[2*MISSION_LEN];
struct fifo_input input;
struct process_acknowledge buffer;
enum control_mode codedmode;
int r;
int essai;    /* number of synchronization attempts */

/* communication init */
   if ( !apiinit )
   {
	wdx_api_init();
	apiinit++;
   }
/* check process rights */
/*   if(!spc_super_user())
/*   {
/*      wdxerrno = WDX_ACCES;
/*      return(-1);
/*   }
*/
/* check pointer validity */
   if(!missionname)
   {
      wdxerrno = WDX_INVAL;
      return(-1);
   }

/* check string length */
   if ( strlen(missionname) >= MISSION_LEN )
   {
      wdxerrno = WDX_INVAL;
      return(-1);
   }

/* check mode validity */
   switch(mode)
   {
   case REQUEST_ONLY :
	codedmode = REQUEST_CONTROL_ONLY;
        break;
   case EVENT_ONLY :
	codedmode = EVENT_CONTROL_ONLY;
        break;
   case REQUEST_AND_EVENT :
	codedmode = REQUEST_AND_EVENT_CONTROL;
        break;
	break;
   default :
	wdxerrno = WDX_INVAL;
	return(-1);
	break;
   }

/* check if already connected */
   if ( (slot = find_slot_index(missionname)) != NOT_AN_INDEX )
   {
	wdxerrno = WDX_EXIST;
	return(-1);
   }
/* check if available slot */
   if ( (slot = find_first_free_slot()) == NOT_AN_INDEX )
   {
        wdxerrno = WDX_NOBUFS;
	return(-1);
   }

/* initialize slot entry */
   reset_application_link(&(wdxtable[slot]));


/* check if server process running */
   pt = &(wdxtable[slot].data);
   if ( lib_get_process_info(missionname,pt) == -1 )
   {
/*printf("lib_get_process_info 1\n"); */
      wdxerrno = WDX_NOTRUN;	/* server is not running */
      return(-1);
   }

/* open communication link */
   if ( (socket = com_open_socket(pt->udp,
			      pt->redundancy,
			      pt->links,
			      WDX_APPLICATION)) == -1 )
   {
      wdxerrno = WDX_NOTCONN;		/* connection refused */
      return(-1);
   }


/* open FIFO */
   com_create_fifo_name(fifoname,missionname);
   if ( ( fifo = com_open_fifo(fifoname,WDX_APPLICATION)) == -1 )
   {
      wdxerrno = WDX_IO;
      com_close_socket(socket);
      return(-1);
   }

/* set wdxtable slot */
   wdxtable[slot].use = API_SLOT_USED;
   strcpy(wdxtable[slot].mission,missionname);
   wdxtable[slot].mode = codedmode;
   wdxtable[slot].socket = socket;
   wdxtable[slot].fifo   = fifo;

/* set the local lock and look at the daemon one */
   if ( spc_set_lock(fifoname) == -1)
   {
      wdxerrno = WDX_SLCK;
      com_close_fifo(fifo);
      com_close_socket(socket);
      return(-1);
   }

   sprintf(daemonlock,"%s%d",missionname,pt->pid);
   if ( spc_lookat_daemon_lock(daemonlock,wdx_ast,slot) != 0)
   {
      wdxerrno = WDX_LKLCK;
      com_close_fifo(fifo);
      com_close_socket(socket);
      return(-1);
   }


/* send OPEN_LINK message */
   lib_forward_operation(socket,
		     WDX_OPEN_CONTROL,
		     wdxtable[slot].identifier,
		     fifoname,
		     codedmode);



/* read connection acknowledge */
   input.descriptor = fifo;
   input.buffer = (char *) &buffer;
   input.size = sizeof(struct process_acknowledge);
/* if the read returns 0 which is end of file then check if the wdx process
   is still running and re-issue the read : done for process synchronization */
   r = 0;
   essai = 0;
   do
   {
       if ((r = com_get_fifo_s_message(&input)) == -1 )
       {
			com_close_fifo(fifo);
	    	com_close_socket(socket);
	    	wdxerrno = WDX_IO;
            return(-1);
       }

	   if (r == 0)
       {
            if ( lib_get_process_info(missionname,&tmp) == -1 )
	    	{
       			com_close_fifo(fifo);
				com_close_socket(socket);
	         	wdxerrno = WDX_NOTRUN;
                return(-1);
	    	}
	    	else
            {
	        	essai++;
		 		if ( essai == SYNCHRO_ATTEMPT )
		 		{
					com_close_fifo(fifo);
	            	com_close_socket(socket);
	            	wdxerrno = WDX_NOTRUN;
                    return(-1);
		 		}
/* MK001 start   else WDX_SLEEP(SYNCHRO_WAIT); */
                else spc_sleep();
/* MK001 end */

       		}
       }
   }
   while(r != input.size);


/* analyze status returned by wdx process */
   if ( buffer.status != PROCESS_SUCCESS )
   {
        com_close_fifo(fifo);
        com_close_socket(socket);
        wdxerrno = WDX_NOBUFS;
        return(-1);
   }

/* return success */
   return(slot);
}

/* wdx_close()
 * Close wdx client environment for the specified descriptor
 *
 * Argument : link connection descriptor
 *
 * Return   : 0 on success or -1 on error
 *
 * Error    : on error wdxerrno is set to
 *		WDX_ACCES if process is not in superuser group
 *		WDX_INVAL if descriptor is invalid
 *
 * Return   : 0 on success, -1 on error
 *
 */
int wdx_close(descriptor)
int descriptor;
{
int slot = NOT_AN_INDEX;
struct fifo_input input;
struct process_acknowledge buffer;
int ret;

/* communication init */
   if ( !apiinit )
   {
	wdx_api_init();
	apiinit++;
   }

/* check process rights */
/*   if(!spc_super_user())
/*   {
/*      wdxerrno = WDX_ACCES;
/*      return(-1);
/*   }
*/
/* check if descriptor available */
   if ( (slot = find_slot_in_use(descriptor)) == NOT_AN_INDEX )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* update process_info */
   if ( lib_get_process_info(wdxtable[slot].mission,&(wdxtable[slot].data)) == -1 )
   {
      wdxerrno = WDX_IO;	/* server is not running */
      ret = -1;
   }

/* send CLOSE_LINK message */
   else
   {
       lib_forward_operation(wdxtable[slot].socket,
		         WDX_CLOSE_CONTROL,
  		         ++wdxtable[slot].identifier,
		         NULL,
		         NOT_A_MODE);

/* get acknowledge */
       input.descriptor = wdxtable[slot].fifo;
       input.buffer = (char *) &buffer;
       input.size = sizeof(struct process_acknowledge);
       if ( com_get_fifo_s_message(&input) == -1 )
       {
        com_close_fifo(wdxtable[slot].fifo);
	com_close_socket(wdxtable[slot].socket);
	wdxerrno = WDX_IO;
       }
       ret = 0;
   }

/* close socket */
   com_close_socket(wdxtable[slot].socket);
/* close fifo */
   com_close_fifo(wdxtable[slot].fifo);

/* free table link */
   reset_application_link(&(wdxtable[slot]));
   return(ret);
}

/* wdx_write()
 * Write a message to the server specified by descriptor. The message is
 * formatted using type and string data.
 *
 * Argument : connection descriptor
 *	      request operation type
 *	      request operation string
 *	      request operation integer
 *	      pointer to request status
 *	      pointer to request acknowledge string
 *	      size of request string
 *
 * Return   : >=0 ( request identifier ) on success or -1 on error
 *
 * Error    : on error wdxerrno is set to
 *		WDX_ACCES if process is not in superuser group
 *		WDX_INVAL if invalid parameter
 *		WDX_IO if I/O erro while writing
 */
int wdx_write(descriptor,type,string,arg,status,message,size)
int descriptor;
int type;
char *string;
int arg;
int *status;
char *message;
int size;
{
int slot = NOT_AN_INDEX;
struct process_acknowledge buffer; /* process acknowledge */
struct fifo_input input;    /* input for acknowledge */

/* communication init */
   if ( !apiinit )
   {
	wdx_api_init();
	apiinit++;
   }

/* check process rights */
/*   if(!spc_super_user())
/*   {
/*      wdxerrno = WDX_ACCES;
/*      return(-1);
/*   }
*/
/* check if descriptor available */
   if ( (slot = find_slot_in_use(descriptor)) == NOT_AN_INDEX )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check if data string available */
   if ( !string )
   {
       wdxerrno = WDX_INVAL;
       return(-1);
   }

/* check if data string available */
   if ( !message )
   {
       wdxerrno = WDX_INVAL;
       return(-1);
   }

/* check if data string not too long */
   if ( strlen(string) >= RESOURCE_LEN )
   {
       wdxerrno = WDX_INVAL;
       return(-1);
   }

/* check type validity */
   if ( check_type(type) == -1 )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* update process_info */
   if ( lib_get_process_info(wdxtable[slot].mission,&(wdxtable[slot].data)) == -1 )
   {
      wdxerrno = WDX_IO;	/* server is not running */
      return(-1);
   }

/* check for master_substitute capabilities */
   if ( type == WDX_MASTER || type == WDX_NOMASTER || type == WDX_GETMASTER )
   {
	if(wdxtable[slot].data.type != MASTER_SUBSTITUTE_T )
	{
	    wdxerrno = WDX_PERM;
	    return(-1);
	}
   }

/* send message */
   lib_forward_operation(wdxtable[slot].socket,
                     type,
                     ++wdxtable[slot].identifier,
                     string,
                     arg);

/* get acknowledge */
   input.descriptor = wdxtable[slot].fifo;
   input.buffer = (char *) &buffer;
   input.size = sizeof(struct process_acknowledge);
   if ( com_get_fifo_s_message(&input) == -1 )
   {
        com_close_fifo(wdxtable[slot].fifo);
	com_close_socket(wdxtable[slot].socket);
	wdxerrno = WDX_IO;
        return(-1);
   }
   if ( buffer.type != PROCESS_ACKNOWLEDGE )
   {
        wdxerrno = WDX_IO;
        return(-1);
   }
   if ( buffer.identifier != wdxtable[slot].identifier )
   {
        wdxerrno = WDX_IO;
        return(-1);
   }
   *status = buffer.status;
   strncpy(message,buffer.data,size);
   return(buffer.identifier);
}

/* wdx_read()
 * Read a message from the FIFO specified by descriptor.
 *
 * Argument : connection descriptor
 *	      pointer to message type
 *	      pointer to message status
 *	      pointer to message identifier
 *	      pointer to event generation number
 *	      pointer to resource status
 *	      pointer to message string
 *	      size of message string
 *
 * Return   : 0 on success or -1 on error
 *
 * Error    : on error wdxerrno is set to
 *		WDX_ACCES if process is not in superuser group
 *		WDX_INVAL if invalid parameter
 *		WDX_IO if I/O erro while writing
 */
int wdx_read(descriptor,type,status,identifier,generation,rstatus,message,length)
int descriptor;
int *type;
int *status;
int *identifier;
int *generation;
int *rstatus;
char *message;
int length;
{
int slot = NOT_AN_INDEX;
struct process_message buffer;
struct fifo_input input;

/* communication init */
   if ( !apiinit )
   {
	wdx_api_init();
	apiinit++;
   }

/* check process rights */
/*   if(!spc_super_user())
/*   {
/*      wdxerrno = WDX_ACCES;
/*      return(-1);
/*   }
*/
/* check if descriptor available */
   if ( (slot = find_slot_in_use(descriptor)) == NOT_AN_INDEX )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check status address validity */
   if ( !status )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check type address validity */
   if ( !type )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check identifier address validity */
   if ( !identifier )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check rstatus address validity */
   if ( !rstatus )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check generation address validity */
   if ( !generation )
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* check message address validity */
   if ( !message)
   {
	wdxerrno = WDX_INVAL;
	return(-1);
   }

/* update process_info */
   if ( lib_get_process_info(wdxtable[slot].mission,&(wdxtable[slot].data)) == -1 )
   {
      wdxerrno = WDX_IO;	/* server is not running */
      return(-1);
   }

/* read message */
   input.descriptor = wdxtable[slot].fifo;
   input.buffer = (char *) &buffer;
   input.size = sizeof(struct process_message);
   if ( com_get_fifo_a_message(&input) == -1 )
   {
        com_close_fifo(wdxtable[slot].fifo);
	com_close_socket(wdxtable[slot].socket);
	wdxerrno = WDX_IO;
        return(-1);
   }
   *type = buffer.type;
   *status = buffer.status;
   *identifier = buffer.identifier;
   *generation = buffer.generation;
   *rstatus = buffer.rstatus;
   strncpy(message,buffer.data,length);
   return(0);
}
