/*
 *   WDX: WitchDoctor-WDX, versatile distributed lock manager
 *   Copyright (C) 2003 Dominique Chabord,
 *   Please refer to credits section below
 *
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ************************************************
 *
 *   History and credits:
 *   First version of WitchDoctor-WDX was based on:
 *
 *   DECwatchdog/Autopilot, watchdog and lock manager
 *   Copyright (C) 2003 HP
 *
 *   Developed by
 *     Computer Special Systems, Annecy France
 *     A Division of Digital Equipment Corporation
 *   and
 *     Custom-Systems, Annecy France
 *     An engineering center of Compaq
 *
 *   Provided under GNU-GPL terms by HP, 30 June 2003
 *   GNU-GPL replaces previous license statements
 *
 *   Rebranded, ported and modified for Shaman-X project, 15 July 2003
 *   by Dominique Chabord
 *   and maintained by other contributors as mentionned in the code
 *
 *   ************************************************
 *
 *   WDX is promoted by Shaman-X project
 *
 *   Check for new versions and documentation at
 *   http://www.shaman-x.org
 *
 *   ************************************************
 *
 *   Most of WDX code is designed to run unchanged on
 *   different operating systems, including at V1.0 time:
 *   Gnu-Linux, WindowsNT, DigitalUNIX and OpenVMS.
 *   Operating system specifics are identified by dedicated files.
 *
 *   Evolutions
 *   ----------
 *   July-2003     Dominique Chabord    WitchDoctor-WDX V1.0
 *                 Creation
 *
 *   ************************************************
 *
 *
 *	wdx_config.c
 *
 *	Configuration file management
 *
 *      Private Routines :
 *
 *      int read_mission()
 *      int read_local()
 *      int read_support()
 *      int script_funct()
 *      int netmask_funct()
 *      int membmask_funct()
 *      int resmask_funct()
 *      int end_funct()
 *      int field_copy()
 *      int name_funct()
 *      int type_funct()
 *	int priority_funct()
 *      int broadcast_funct()
 *      int validity_funct()
 *      int redundancy_funct()
 *      int network1_funct()
 *      int network2_funct()
 *      int logflag_funct()
 *      int logfile_funct()
 *      int period_funct()
 *      int priority_funct()
 *      int activity_funct()
 *      int threshold_funct()
 *      int statistics_funct()
 *      int is_numberic()
 *	void encode_mission_file()
 *	void decode_mission_file()
 *	int conf_checksum()
 *	int unique_key()
 *
 *	Public Routines :
 *
 *      int cfg_config_init()
 ************************************************************************/
/*
 *
 *  11-Feb-94 -- Cornu
 *      Creation
 */

/*
 * Declaration
 */
#ifndef WDX_CONFIG_MODULE
#define WDX_CONFIG_MODULE
#endif  /* WDX_CONFIG_MODULE */

#include	"wdx.h"
#include	"wdx_config.h"

#define		LEN	STRING_LEN	/* long line		*/

void clear_leading_char(char *c);
int yesno(struct mask_element *c,char *w,char *k);
int get_version(char *s,unsigned char *h,unsigned char	*l);
int get_date(char *s,char *r);

/*
 * PUBLIC ROUTINES : prefixe is "cfg"
 */

/* int cfg_config_init()
 * Read the configuration files :
 *	- mission file
 *	- local file
 *	- support file
 * Set the group data base.
 *
 * Argument : pt to mission name
 *            pt to group data base
 *
 * Return   : 0 ok / -1 fatal error
 */
int cfg_config_init(missionpt,g)
char          *missionpt;	    /* mission name pointer */
struct group          *g;	    /* group data structure pointer */
{
int                _configcount;    /* number of members in configuration */
char          _rcfile[PATH_MAX];    /* file name */
struct mission_file missionimage;   /* mission file image */
struct local_file     localimage;   /* local file image */
struct support_file supportimage;   /* suppport image */
struct mission_file        *_fm;    /* pt to image of the mission file */
struct local_file          *_fl;    /* pt to image of the local file */
struct support_file        *_fs;    /* pt to image of the support file */
struct member               *_p;    /* member pt */
struct interface_char       *_n;    /* interface pt */
struct local_line           *_o;    /* pt to localdes table element */
int                          _i;    /* index */
int			     _k;    /* index */
enum networks                _j;    /* network index */

    /* verify that mission is defined */
    if ( missionpt == NULL || g == NULL )
    {
         PERROR_B
         "Cfg_config_init parameter error"
         PERROR_E;
         return(-1);
    }

    /* initialize wdxversion */
    g->wdxversion = WDX_VERSION;

    /* initialize group data base */
    g->missionkey = unique_key(missionpt);
    strncpy(g->mission,missionpt,MISSION_LEN);

    _configcount = 0;

    /* define mission file name */
    sprintf(_rcfile,MISSION_FILE,wdxdirectory,missionpt,missionpt);
    memset(&missionimage,'\0',sizeof(struct mission_file));

    /* read mission file */
    if ( read_mission(_rcfile,&missionimage,&_configcount) != 0 )
         return(-1);

    /* define local file name and set default values */
    sprintf(_rcfile,LOCAL_FILE,wdxdirectory,missionpt,missionpt);
    memset(&localimage,'\0',sizeof(struct local_file));
    for(_o=(localdes+EVENTSCRIPT_MASK_BEGIN);_o->keyword;_o++)
        if( _o->funct )
	    (*_o->funct)(&localimage,
                         _o->index1,
                         _o->index2,
                         NO_STRING,
                         _o->keyword);
/*    for(_i=0;_i<MEMBER_STATUS;_i++)
	for(_j=0;_j<MEMBER_STATUS;_j++)
	    localimage.membermasks[_i][_j] = NO;
    for(_i=0;_i<NETWORK_STATUS;_i++)
	for(_j=0;_j<NETWORK_STATUS;_j++)
	    localimage.networkmasks[_i][_j] = NO;
    for(_i=0;_i<RESOURCE_STATUS;_i++)
	for(_j=0;_j<RESOURCE_STATUS;_j++)
	    localimage.resourcemasks[_i][_j] = NO;
    localimage.mismatchmask = NO;
*/
    /* read local file */
    if ( read_local(_rcfile,&localimage) == -1 )
         return(-1);

    /* define support file name and set default values */
    sprintf(_rcfile,SUPPORT_FILE,wdxdirectory,missionpt,missionpt);
    supportimage.log = ENABLE;
    sprintf(supportimage.logfile,LOG_FILE,wdxdirectory,missionpt,missionpt);
    supportimage.period = PERIOD_VAL;
    supportimage.activity = DISABLE;
    supportimage.priority = HIGHEST;
    supportimage.statistics = DISABLE;
    supportimage.thresholdcount = 0;
    for(_i=0;_i<MEMBER_COUNT;_i++)
	supportimage.thresholdtable[_i].index = NOT_AN_INDEX;

    /* read mission file */
    if ( read_support(_rcfile,&supportimage,&missionimage,_configcount) == -1 )
         return(-1);

	/* read the members IP addresses file */
    sprintf(_rcfile,MEMBER_FILE,wdxdirectory,missionpt,missionpt);
	_i = read_members_ip (_rcfile);		/* see if any IP addresses defined */
	if (_i == -1)
		return (-1);		/* invalid */
	if (_i == 0)
		using_ip_addresses = 0;		/* if no IP addresses defined */

    /* group init */
    _fl = &localimage;
    /* script init */
    strcpy(g->nosubstitutescript,_fl->script[NOSUBSTITUTESCRIPT_INDEX]);
    strcpy(g->newmasterscript,_fl->script[NEWMASTERSCRIPT_INDEX]);
    strcpy(g->satellitescript,_fl->script[SATELLITESCRIPT_INDEX]);
    strcpy(g->exceptionscript,_fl->script[EXCEPTIONSCRIPT_INDEX]);
    strcpy(g->eventscript,_fl->script[EVENTSCRIPT_INDEX]);
/*
//printf("master : %s\n",g->nosubstitutescript);
//printf("substitute : %s\n",g->newmasterscript);
//printf("satellite : %s\n",g->satellitescript);
//printf("exception : %s\n",g->exceptionscript);
//printf("event : %s\n",g->eventscript);
*/
    /* member table init */
    g->membercount = _configcount;
    _fm = &missionimage;
    for(_i=0;_i<VERSION_WORDS_COUNT;_i++)
         g->version[_i] = _fm->version[_i];
    strncpy(g->date,_fm->date,DATE_LEN);
    for (_i=0;_i<g->membercount;_i++)
    {
         _p = &(g->memberlist[_i]);
         strcpy(_p->name,_fm->memberlist[_i].name);
         _p->type = _fm->memberlist[_i].type;
         _p->broadcast = _fm->memberlist[_i].broadcast;
         _p->validity = _fm->memberlist[_i].validity;
         _p->redundancy = _fm->memberlist[_i].redundancy;
	 _p->priority = _fm->memberlist[_i].priority;
         for(_j = NETWORK_1;_j<NETWORKS;_j++)
         {
             _n = &(_p->networks[_j]);
             strcpy(_n->ifname,_fm->memberlist[_i].netname[_j]);
             if ( spc_strcasecmp(_fm->memberlist[_i].netname[_j],
                  NETWORK_NA) != 0 )
             {
                  _n->use = USED;
                  _n->netnb = _j;
             }
             else
             {
                  _n->use = NOTUSED;
                  _n->netnb = _j;
             }
	     _n->threshold = DISABLE;
             _n->value    = _p->validity -1;
             _n->count    = 0;
             _n->max      = 0;
         }
    }

    /* Store script mask information into group data base */
    _fl = &localimage;
    for(_i=0;_i<MEMBER_STATUS;_i++)
	for(_j=0;_j<MEMBER_STATUS;_j++)
    {
	    g->membermasks[_i][_j].value = _fl->membermasks[_i][_j].value;
	    g->membermasks[_i][_j].key = _fl->membermasks[_i][_j].key;
    }
    for(_i=0;_i<NETWORK_STATUS;_i++)
	for(_j=0;_j<NETWORK_STATUS;_j++)
    {
	    g->networkmasks[_i][_j].value = _fl->networkmasks[_i][_j].value;
	    g->networkmasks[_i][_j].key = _fl->networkmasks[_i][_j].key;
    }
    for(_i=0;_i<RESOURCE_STATUS;_i++)
	for(_j=0;_j<RESOURCE_STATUS;_j++)
    {
	    g->resourcemasks[_i][_j].value = _fl->resourcemasks[_i][_j].value;
	    g->resourcemasks[_i][_j].key = _fl->resourcemasks[_i][_j].key;
    }
    g->mismatchmask.value = _fl->mismatchmask.value;
    g->mismatchmask.key = _fl->mismatchmask.key;

    /* print information */
/*
    printf("member masks \n");
    for(_i=0;_i<MEMBER_STATUS;_i++)
    {
	for(_j=0;_j<MEMBER_STATUS;_j++)
	    printf("\t %s %d",g->membermasks[_i][_j].key,
                              g->membermasks[_i][_j].value);
	printf("\n");
    }
    printf("network masks \n");
    for(_i=0;_i<NETWORK_STATUS;_i++)
    {
	for(_j=0;_j<NETWORK_STATUS;_j++)
	    printf("\t %s %d",g->networkmasks[_i][_j].key,
                              g->networkmasks[_i][_j].value);
	printf("\n");
    }
    printf("resource masks \n");
    for(_i=0;_i<RESOURCE_STATUS;_i++)
    {
	for(_j=0;_j<RESOURCE_STATUS;_j++)
	    printf("\t %s %d",g->resourcemasks[_i][_j].key,
			      g->resourcemasks[_i][_j].value);
	printf("\n");
    }
    printf("mismatch mask \n");
    printf("\t %s %d\n",g->mismatchmask.key,g->mismatchmask.value);
*/
    /* Store support information in group data base */
    _fs = &supportimage;
    g->log = _fs->log;
    strcpy(g->logfile,_fs->logfile);
    g->priority = _fs->priority;
    g->period   = _fs->period;
    g->activity = _fs->activity;
    g->statistics = _fs->statistics;
    g->statisticsperiod = _fs->statisticsperiod;
/*
    printf(" log : %d\n",g->log);
    printf("%s\n",g->logfile);
    printf("priority : %d\n",g->priority);
    printf("activity : %d\n",g->activity);
    printf("period : %d\n",g->period);
    printf("statistics : %d\n",g->statistics);
    printf("statistics period : %d\n",g->statisticsperiod);
*/
    /* Store threshold information in group data base */
    _fs = &supportimage;
    for(_i=0;_i<_fs->thresholdcount;_i++)
	for(_j=0;_j<NETWORKS;_j++)
	{
	    _k = _fs->thresholdtable[_i].index;
	    g->memberlist[_k].networks[_j].value = _fs->thresholdtable[_i].value;
	    g->memberlist[_k].networks[_j].threshold = ENABLE;
	}

    /* calculate config file checksum */
    if ( conf_checksum(_fm,&(g->configversion)) == -1 )
         return(-1);
/*
for(_i=0;_i<_configcount;_i++)
    {
    printf("%s %d %d %d %d\n",
           g->memberlist[_i].name,
           g->memberlist[_i].type,
           g->memberlist[_i].broadcast,
           g->memberlist[_i].validity,
           g->memberlist[_i].redundancy);
    printf("%s %d %d %d %d %d \n",
           g->memberlist[_i].networks[NETWORK_1].ifname,
           g->memberlist[_i].networks[NETWORK_1].netnb,
           g->memberlist[_i].networks[NETWORK_1].use,
           g->memberlist[_i].networks[NETWORK_1].status,
           g->memberlist[_i].networks[NETWORK_1].threshold,
           g->memberlist[_i].networks[NETWORK_1].value);
    printf("%s %d %d %d %d %d\n",
           g->memberlist[_i].networks[NETWORK_2].ifname,
           g->memberlist[_i].networks[NETWORK_2].netnb,
           g->memberlist[_i].networks[NETWORK_2].use,
           g->memberlist[_i].networks[NETWORK_1].status,
           g->memberlist[_i].networks[NETWORK_1].threshold,
           g->memberlist[_i].networks[NETWORK_1].value);
    }
*/
    return(0);
}

/*
 * LOCAL ROUTINES
 */

/* int read_mission()
 * Read the mission file and analyze each line. The results are
 * stored in a mission_file structure.
 * Leading and trailing white spaces are removed into the analyze functions
 * which are
 * version_line_analyze(), date_line_analyze() and member_line_analyze()
 *
 * Arguments : pointer to mission file name
 *             pointer to mission_file structure
 *             pointer to count of member
 *
 * Return    : -1 if error
 *             0 if OK
 */
int read_mission(filename,c,countpt)
char        *filename;       /* pointer to mission file name */
struct mission_file *c;	     /* structure to fill in	*/
int          *countpt;       /* pointer to count of member */
{
FILE        *_fp;            /* pointer to file	*/
char   _buf[LEN];
int       _words;
char    _w0[LEN];
char    _w1[LEN];
char         *_s;
int         _ret;
int     _versioncount;	    /* number of version lines */
int     _datecount;         /* number of date lines */
int     _computed;	    /* flag to detect invalid keywords */
struct file_analyze *_k;

    _datecount    = 0;
    _versioncount = 0;
    *countpt      = 0;
    missionkeydes[VERSION_ELEMENT].cpt = &_versioncount;
    missionkeydes[DATE_ELEMENT].cpt = &_datecount;
    missionkeydes[MEMBER_ELEMENT].cpt = countpt;

    if ((_fp=fopen(filename,"r"))== (FILE *)NULL)
    {
        sprintf(messagestring,"Can't open mission file %s",filename);
        PERROR_B
        messagestring
        PERROR_E;
        return(-1);
    }

    while(fgets(_buf,LEN,_fp)!=NULL)
    {
//printf("toto\n"); fflush(stdout);
	/* now scan all lines */
	for(_s=_buf;isspace(*_s);_s++);	/* skip white spaces 	*/
//printf("toto1\n"); fflush(stdout);
	if (*_s == '\0') continue;	/* skip empty lines  	*/
//printf("toto2\n"); fflush(stdout);
//printf("%s\n",_s); fflush(stdout);
	if (*_s == '#') continue;	/* skip comment lines 	*/
//printf("toto3\n"); fflush(stdout);
        _w0[0] = '\0';
        _w1[0] = '\0';
        _words=sscanf(_s,"%s%[^\n]",_w0,_w1);
//printf("--> %s et %s\n", _w0, _w1); fflush(stdout);
        _computed = 0;

	for(_k=missionkeydes;_k->keyword;_k++)
	    if(spc_strcasecmp(_w0,_k->keyword) == 0)
	{
	    _computed++;
	    _ret = (*_k->funct)(_w1,c,_k->cpt);
	}
        if ( _computed == 0 )
	{
	    sprintf(messagestring,
		    "Invalid keyword (%s)",_w0);
	    _ret = -1;
	}
        if ( _ret == -1 )
	{
	    PERROR_B
	    "Mission file syntax"
	    PERROR_E;
	    PCOMPLEMENT_B
	    messagestring
	    PCOMPLEMENT_E;
	    PCOMPLEMENT_B
            _s
            PCOMPLEMENT_E;
	    fclose(_fp);
	    return(-1);
	}
    }
    fclose(_fp);
    if ( _versioncount != VERSION_COUNT )
    {
	PERROR_B
	"Missing version line in mission file "
	PERROR_E;
	return(-1);
    }
    if ( _datecount != DATE_COUNT )
    {
	PERROR_B
	"Missing date line in mission file "
	PERROR_E;
	return(-1);
    }
    if ( *countpt == 0 )
    {
	PERROR_B
	"Missing member line in mission file"
	PERROR_E;
	return(-1);
    }
    return(0);
}

/* int version_line_analyze()
 * Analyze the version line field
 *
 * Argument : pt to read line
 *	      pt to mission file structure
 *	      pt to version line count
 *
 * Return : 0 ok / -1 error
 */
int version_line_analyze(s,c,p)
char                *s;	/* line read without keyword */
struct mission_file *c; /* structure to fill in	*/
int                 *p; /* pointer to version line count */
{
char *_s; /* pt to line without space */

   if ( *p == VERSION_COUNT )
   {
        PWARNING_B
        "Too many versions listed in mission file"
        PWARNING_E;
        return(0);
   }
   else
   {
	for(_s=s;isspace(*_s);_s++);	/* skip white spaces 	*/
        clear_leading_char(_s);
	if ( get_version(_s,&(c->version[0]),&(c->version[1])) == -1 )
	     return(-1);
	else (*p)++;
   }
	return(0);
}

/* int date_line_analyze()
 * Analyze one date line
 *
 * Argument : pt to read line
 *	      pt to mission file structure
 *	      pt to date count
 *
 * Return : 0 ok / -1 error
 */
int date_line_analyze(s,c,p)
char                *s;	/* line read without keyword */
struct mission_file *c; /* structure to fill in	*/
int                 *p; /* pointer to date count line */
{
char               *_s; /* pt to effective date string */

   if ( *p == DATE_COUNT )
   {
        PWARNING_B
        "Too many dates listed in mission file"
        PWARNING_E;
        return(0);
   }
   else
   {
	for(_s=s;isspace(*_s);_s++);	/* skip white spaces 	*/
        clear_leading_char(_s);
	if ( get_date(_s,c->date) == -1 )
	     return(-1);
        (*p)++;
   }
   return(0);
}

/* int member_line_analyze()
 * Analyze one member line field after field
 *
 * Argument : pt to read line
 *	      pt to mission file structure
 *	      pt to member count
 *
 * Return : 0 ok / -1 error
 */
int member_line_analyze(s,c,p)
char                *s;	/* line read without keyword */
struct mission_file *c; /* structure to fill in	*/
int                 *p; /* pointer to count of member */
{
int  _ret;
int  _i;
int  _words;
char _w[MEMBER_COUNT_WORDS][LEN];
char *_s; /* pt to effective date string */
char *_p; /* pt to effective field */

   for(_i=0;_i<MEMBER_COUNT_WORDS;_i++)
        _w[_i][0] = '\0';
   if ( *p == MEMBER_COUNT )
   {
        PWARNING_B
        "Too many nodes listed in mission file"
        PWARNING_E;
        return(0);
   }
   else
   {
	for(_s=s;isspace(*_s);_s++);	/* skip white spaces 	*/
        _words = sscanf(_s,"%s %s %s %s %s %s %s%[^\n]",
                        _w[0],_w[1],_w[2],_w[3],_w[4],_w[5],_w[6], _w[7]);
        if ( _words != MEMBER_COUNT_WORDS )
        {
             sprintf(messagestring,
             "Invalid mission file format for line \n%s",_s);
             return(-1);
        }
	_ret = 0;
	_i = 0;
        while(_i<MEMBER_COUNT_WORDS && _ret>=0)
        {
            for(_p=_w[_i];isspace(*_p);_p++);     /* skip white space and tab */
            clear_leading_char(_p);
             _ret = (memberdes[_i].funct)(&(c->memberlist[*p]),_p);
	     _i++;
	}
	if ( _ret == 0 )
	{
	     for(_i=0;_i<(*p -1);_i++)
	   	 if(spc_strcasecmp(c->memberlist[_i].name,
                          c->memberlist[*p].name) == 0 )
             {
	  	  _ret = -1;
	  	  sprintf(messagestring, "Member %s registered twice",
	 	  c->memberlist[*p].name);
             }
	}
	if ( _ret != 0 )
	{
             return(-1);
        }
        (*p)++;
    }
    return(0);
}

/* int read_local()
 * Read the local file and analyze each line. The results are
 * stored in a local_file structure.
 * Leading and trailing spaces are removed in this routine
 *  because all lines have the same format which is one keyword and one value.
 *
 *
 * Arguments : pointer to local file name
 *             pointer to local_file structure
 *
 * Return    : -1 if error
 *             0 if OK
 */
int read_local(filename,c)
char        *filename;       /* pointer to local file name */
struct local_file  *c;	     /* structure to fill in	*/
{
FILE        *_fp;            /* pointer to file	*/
char   _buf[LEN];	     /* buffer in file */
char    _w0[LEN];	     /* keyword */
char    _w1[LEN];	     /* end of the line */
int             _words;	     /* nb of words in line */
char               *_s;	     /* pt */
char               *_t;	     /* pt */
struct local_line  *_p;	     /* pt */
int		  _ret;      /* returned value */

    if ((_fp=fopen(filename,"r"))== (FILE *)NULL)
    {
        sprintf(messagestring,"Can't open local file %s",filename);
        PERROR_B
        messagestring
        PERROR_E;
        return(-1);
    }

    while(fgets(_buf,LEN,_fp)!=NULL)
    {
	/* now scan all lines */
	for(_s=_buf;isspace(*_s);_s++);	/* skip white spaces 	*/
	if (*_s == '\0') continue;	/* skip empty lines  	*/
	if (*_s == '#') continue;	/* skip comment lines 	*/
        _w0[0] = '\0';
        _w1[0] = '\0';
	_words=sscanf(_s,"%s%[^\n]",_w0,_w1);
	for(_t=_w1;isspace(*_t);_t++);	/* skip white spaces 	*/

        if ( _words < LOCAL_COUNT_WORDS )
        {
             sprintf(messagestring,
                     "Invalid local file format for line \n%s",
                     _s);
             PERROR_B
             messagestring
             PERROR_E;
             fclose(_fp);
             return(-1);
        }
        for(_p=localdes; _p->funct; _p++)
	{
	     _ret = -2;
	     if ( _p->keyword == 0 )
	     {
	          sprintf(messagestring,"Invalid keyword (%s)",_w0);
		  _ret = -1;
	     }
	     else if ( spc_strcasecmp(_w0,_p->keyword) == 0 )
                  _ret = (*_p->funct)(c,_p->index1,_p->index2,_t,_p->keyword);
	     if ( _ret == -1 )
             {
                  PERROR_B
		  "Local file syntax"
		  PERROR_E;
		  PCOMPLEMENT_B
		  messagestring
		  PCOMPLEMENT_E;
		  PCOMPLEMENT_B
                  _s
                  PCOMPLEMENT_E;
                  fclose(_fp);
                  return(-1);
             }
	     if ( _ret == 0 )
		  break;
        }
    }
    fclose(_fp);
    return(0);
}

/* int read_support()
 * Read the support file and analyze each line. The results are
 * stored in a support_file structure.
 * Support file is an optional file, so error on open is not fatal.
 * However, syntax error if file exist is fatal.
 *
 * Leading white space are removed in each functional routine because lines
 * does not have the same format.
 *
 * If threshold is defined for a member, this routine checks in the mission
 * file is the member is referenced.
 *
 * Arguments : pointer to support file name
 *             pointer to support_file structure
 *	       pointer to mission_file structure for coherence checking
 *	       number of members in mission file
 *
 * Return    : -1 if error
 *             0 if OK / 0 if file does not exist
 */
int read_support(filename,c,d,e)
char        *filename;       /* pointer to support file name */
struct support_file *c;	     /* structure to fill in	*/
struct mission_file *d;	     /* structure to check */
int		     e;	     /* nb of members */
{
FILE               *_fp;     /* pointer to file	*/
char          _buf[LEN];     /* read buffer */
char           _w0[LEN];
char           _w1[LEN];
char           _w2[LEN];
int              _words;
char                *_s;
struct support_line *_p;
int		   _ret;     /* returned value */

    if ((_fp=fopen(filename,"r"))== (FILE *)NULL)
    {
        return(0);
    }
    while(fgets(_buf,LEN,_fp)!=NULL)
        {
	/* now scan all lines */
	for(_s=_buf;isspace(*_s);_s++);	/* skip white spaces 	*/
	if (*_s == '\0') continue;	/* skip empty lines  	*/
	if (*_s == '#') continue;	/* skip comment lines 	*/
        _w0[0] = '\0';
        _w1[0] = '\0';
        _w2[0] = '\0';
	_words=sscanf(_s,"%s%[^\n]",_w0,_w1);

        if ( _words < SUPPORT_COUNT_WORDS )
        {
             sprintf(messagestring,
                     "Invalid support file format for line \n%s",
                     _s);
             PERROR_B
             messagestring
             PERROR_E;
             fclose(_fp);
             return(-1);
        }
        for(_p=supportdes; _p->funct; _p++)
	{
	     _ret = -2;
             if ( _p->keyword == 0 )
	     {
	          sprintf(messagestring,"Invalid keyword (%s)",_w0);
		  _ret = -1;
	     }
             else if ( spc_strcasecmp(_w0,_p->keyword) == 0 )
		  _ret = (*_p->funct)(c,_w1,d,e);
             if ( _ret == -1 )
             {
                  PERROR_B
		  "Support file syntax"
		  PERROR_E;
		  PCOMPLEMENT_B
		  messagestring
		  PCOMPLEMENT_E;
		  PCOMPLEMENT_B
                  _s
                  PCOMPLEMENT_E;
                  fclose(_fp);
                  return(-1);
             }
	     if ( _ret == 0 )
		  break;
        }

    }
    fclose(_fp);
    return(0);
}

/* int read_members_ip()
 * Read the members_ip file and analyze each line.
 * File is an optional file, so error on open is not fatal.
 * However, syntax errors if the file exists are fatal.
 *
 * Leading white space is removed in each functional routine because lines
 * does not have the same format.

 * only the first 16 entries are read
 *
 * Arguments : pointer to members file name
 *
 * Return    : -1 if error
 *             0 if file does not exist or no IP addresses
 *				n if file read OK and contains n IP addresses
 */
int read_members_ip (filename)
char        *filename;       /* pointer to members_ip file name */
{
FILE               *_fp;     /* pointer to file	*/
char          _buf[LEN];     /* read buffer */
char           _w0[LEN];
char           _w1[LEN];
int              _words;
char                *_s;
int ifaddr;
int i;

	if ((_fp = fopen(filename,"r")) == (FILE *)NULL)
    {
        return(0);
    }

	for (i = 0; i < MEMBER_COUNT; i++)
		members_ip[i].s_addr = 0;
	i = 0;

	while (fgets(_buf,LEN,_fp) != NULL)
    {										/* now scan all lines */
		for(_s=_buf; isspace(*_s); _s++);	/* skip white spaces 	*/

		if (*_s == '\0')
			continue;	/* skip empty lines  	*/
		if (*_s == '#')
			continue;	/* skip comment lines 	*/

		_w0[0] = '\0';
        _w1[0] = '\0';
		_words = sscanf(_s,"%15s%[^\n]",_w0,_w1);


		if (_words != 0)
		{
			if ((ifaddr = inet_addr (_w0)) == 0)
			{
            	sprintf(messagestring,
                     "Invalid Members IP file format for line \n%s",
                     _s);
             	PERROR_B
             	messagestring
             	PERROR_E;
             	fclose(_fp);
             	return(-1);
			}

			if (i < MEMBER_COUNT)
			{
				members_ip[i++].s_addr = inet_addr (_w0);
			}
		}

	}
    fclose(_fp);
    return(i);
}

int script_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : index of script in script table */
int i2;	    /* index2 : NULL */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
   strncpy(c->script[i1],w,SCRIPT_LEN);
   return(0);
}

int date_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : NULL */
int i2;	    /* index2 : NULL */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
char d[DATE_LEN+1];
   clear_leading_char(w);
   return(get_date(w,d));
}

int version_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : NULL */
int i2;	    /* index2 : NULL */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
unsigned char h;
unsigned char l;

   clear_leading_char(w);
   return(get_version(w,&h,&l));
}

int netmask_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : index of old status in networkmasks table */
int i2;	    /* index2 : index of new status in networkmasks table */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
   clear_leading_char(w);
   return(yesno(&(c->networkmasks[i1][i2]),w,k));
}

int membmask_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : index of old status in membermasks table */
int i2;	    /* index2 : index of new status in membermasks table */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
   clear_leading_char(w);
   return(yesno(&(c->membermasks[i1][i2]),w,k));
}

int resmask_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : index of old status in resourcemasks table */
int i2;	    /* index2 : index of new status in resourcemasks table */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
   clear_leading_char(w);
   return(yesno(&(c->resourcemasks[i1][i2]),w,k));
}

int mismask_funct(c,i1,i2,w,k)
struct local_file *c;
int i1;	    /* index1 : NOT USED */
int i2;	    /* index2 : NOT USED */
char *w;    /* pt to script name */
char *k;    /* pt to keyword string */
{
   clear_leading_char(w);
   return(yesno(&(c->mismatchmask),w,k));
}

int yesno(c,w,k)
struct mask_element *c; /* pt to location of result */
char *w; /* pt to script name */
char *k; /* pt to keyword string */
{
   if ( spc_strcasecmp(w,YES_STRING) == 0 )
   {
        c->value = YES;
        c->key = k;
   }
   else if ( spc_strcasecmp(w,NO_STRING) == 0 )
   {
        c->value = NO;
        c->key = k;
   }
   else
   {
        sprintf(messagestring,"Mask (%s) value must be either %s or %s",
	        w, YES_STRING, NO_STRING);
	return(-1);
   }
   return(0);
}

end_funct()
{
   return(-1);
}

errormember(c,w)
char *c;
char *w;
{
   sprintf(messagestring,
           "Error in member keyword analysis \n%s",
           w);
   return(-1);
}

int name_funct(c,w)
struct member_def *c;
char *w;
{
   strncpy(c->name,w,sizeof(c->name));
   return(0);
}

int type_funct(c,w)
struct member_def *c;
char *w;
{
   if ( spc_strcasecmp(w,MASTER_CANDIDATE_STRING) == 0 )
        c->type = MASTER_SUBSTITUTE_T;
   else if ( spc_strcasecmp(w,SATELLITE_CANDIDATE_STRING) == 0 )
        c->type = SATELLITE_T;
   else {
        sprintf(messagestring,
	        "Type value (%s) must be either %s or %s in line",
		w,
		MASTER_CANDIDATE_STRING,
		SATELLITE_CANDIDATE_STRING);
        return(-1);
        }
   return(0);
}

int priority_funct(c,w)
struct member_def *c;
char *w;
{
int _priorityvalue;

   sprintf(messagestring,
           "Priority value (%s) must be a positive integer value",
           w);
   if ( lib_is_numeric(w) )
   {
        return(-1);
   }
   _priorityvalue = atoi(w);
   if ( _priorityvalue < 0 )
   {
        return(-1);
   }
   c->priority = _priorityvalue;
   return(0);
}

int broadcast_funct(c,w)
struct member_def *c;
char *w;
{
int _brdvalue;

   sprintf(messagestring,
           "Broadcast value (%s) must be a non null positive integer value",
           w);
   if ( lib_is_numeric(w) )
   {
        return(-1);
   }
   _brdvalue = atoi(w);
   if ( _brdvalue <= 0 )
   {
        return(-1);
   }
   c->broadcast = _brdvalue;
   return(0);
}

int validity_funct(c,w)
struct member_def *c;
char *w;
{
int _val;

   if ( lib_is_numeric(w) )
   {
        sprintf(messagestring,
                "Validity value (%s) must be a positive integer value",
                w);
        return(-1);
   }
   _val = atoi(w);
   if ( _val <= c->broadcast)
   {
        sprintf(messagestring,
"Validity value (%d) must be a positive integer, greater than broadcast value (%d)",
                _val,c->broadcast);
        return(-1);
   }
   c->validity = _val;
   return(0);
}

int redundancy_funct(c,w)
struct member_def *c;
char *w;
{
int _val;

   sprintf(messagestring,
           "Redundancy value (%s) must be a positive integer value",
           w);
   if ( lib_is_numeric(w) )
   {
        return(-1);
   }
   _val = atoi(w);
   if ( !_val )
   {
        return(-1);
   }
   c->redundancy = _val;
   return(0);
}

int network1_funct(c,w)
struct member_def *c;
char *w;
{

   strncpy(c->netname[NETWORK_1],w,sizeof(c->netname[NETWORK_1]));
   if ( spc_strcasecmp(c->netname[NETWORK_1],NETWORK_NA) == 0 )
        strcpy(c->netname[NETWORK_1],NETWORK_NA);
   return(0);
}

int network2_funct(c,w)
struct member_def *c;
char *w;
{
   strncpy(c->netname[NETWORK_2],w,sizeof(c->netname[NETWORK_2]));
/*   clear_leading_char(c->netname[NETWORK_2]); */
   if ( spc_strcasecmp(c->netname[NETWORK_2],NETWORK_NA) == 0 )
        strcpy(c->netname[NETWORK_2],NETWORK_NA);
   return(0);
}

int logflag_funct(c,w,d,e)
struct support_file *c;
char                *w;
struct mission_file *d;	/* not used */
int		     e;	/* not used */
{
char *_p;

   for(_p=w;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   if ( spc_strcasecmp(_p,DISABLE_STRING) == 0 )
	c->log = DISABLE;
   else if ( spc_strcasecmp(_p,ENABLE_STRING) == 0 )
        c->log = ENABLE;
   else
   {
	sprintf(messagestring,"Log_flag (%s) must be either %s or %s",
		_p,DISABLE_STRING,ENABLE_STRING);
	return(-1);
   }
   return(0);
}

int logfile_funct(c,w,d,e)
struct support_file *c;
char *w;
struct mission_file *d;	/* not used */
int		     e;	/* not used */
{
char *_p;

   for(_p=w;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   strncpy(c->logfile,_p,sizeof(c->logfile));
   return(0);
}

int period_funct(c,w,d,e)
struct support_file *c;
char *w;
struct mission_file *d;	/* not used */
int		     e;	/* not used */
{
int _periodvalue;
char      *_p;

   for(_p=w;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   sprintf(messagestring,
           "Period value (%s) must be a non null positive integer value",
           _p);
   if ( lib_is_numeric(_p) )
   {
        return(-1);
   }
   _periodvalue = atoi(_p);
   if ( _periodvalue <= 0 )
   {
        return(-1);
   }
   c->period = _periodvalue;
   return(0);
}

int os_priority_funct(c,w,d,e)
struct support_file *c;
char                *w;
struct mission_file *d;	/* not used */
int		     e;	/* not used */
{
char *_p;

/* The code for this function is specific to the OS */
/* Refer to wdx_specific_xxx.c to find the code */
   for(_p=w;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   return(spc_priority_funct(&(c->priority),_p));
}

int activity_funct(c,w,d,e)
struct support_file *c;
char                *w;
struct mission_file *d;	/* not used */
int		     e;	/* not used */
{
char *_p;

   for(_p=w;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   if ( spc_strcasecmp(_p,DISABLE_STRING) == 0 )
	c->activity = DISABLE;
   else if ( spc_strcasecmp(_p,ENABLE_STRING) == 0 )
        c->activity = ENABLE;
   else
   {
	sprintf(messagestring,"Activity (%s) must be either %s or %s",
		_p,DISABLE_STRING,ENABLE_STRING);
	return(-1);
   }
   return(0);
}

int threshold_funct(c,w,d,e)
struct support_file *c;
char *w;
struct mission_file *d;	/* mission_file table */
int		     e;	/* nb of members in mission_file table */
{
char _v1[LEN];
char _v2[LEN];
int  _n;
int  _i;
int  _j;
char *_p;
int  _thresvalue;
static int _toomuch=0;

   if ( (c->thresholdcount == e) && (_toomuch == 0 ))
   {
	_toomuch++;
	sprintf(messagestring,"Too many threshold lines defined");
	PWARNING_B
	"Support file format",
	messagestring
	PWARNING_E;
	return(0);
   }
   _v1[0] = '\0';
   _v2[0] = '\0';
   _n = sscanf(w,"%s%[^\n]",_v1,_v2);
   if ( _n != THRESHOLD_COUNT_WORDS )
   {
	sprintf(messagestring,"Invalid threshold line format");
	return(-1);
   }
   /* v1 represents a member : check if defined in mission table */
   for(_p=_v1;isspace(*_p);_p++);
   _j = -1;
   for(_i=0;_i<e;_i++)
   {
	if(spc_strcasecmp(_p,d->memberlist[_i].name) == 0 )
	{
	    _j = _i;
	    break;
	}
   }
   if ( _j == -1 )
   {
	sprintf(messagestring,
	"Member %s not defined in mission file",_p);
	return(-1);
   }
   /* check if thresholds already defined for this index */
   for(_i=0;_i<c->thresholdcount;_i++)
       if(c->thresholdtable[_i].index == _j )
   {
	sprintf(messagestring,
	"Threshold defined twice for member %s",_p);
	return(-1);
   }

   /* check if value is an integer : value must be >0 and less than broadcast */
   for(_p=_v2;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   sprintf(messagestring,
   "Threshold value (%s) must be a non null positive integer less than validity value (%d)",
           _p,d->memberlist[_j].validity);
   if ( lib_is_numeric(_p) )
   {
        return(-1);
   }
   _thresvalue = atoi(_p);
   if ( _thresvalue <= 0 || _thresvalue >= d->memberlist[_j].validity )
   {
        return(-1);
   }
   c->thresholdtable[c->thresholdcount].index = _j;
   c->thresholdtable[c->thresholdcount].value = _thresvalue;
   c->thresholdcount++;
   return(0);
}

int statistics_funct(c,w,d,e)
struct support_file *c;
char *w;
struct mission_file *d;	/* not used */
int		     e;	/* not used */
{
int _statvalue;
char      *_p;

   for(_p=w;isspace(*_p);_p++);     /* skip white space and tab */
   clear_leading_char(_p);
   sprintf(messagestring,
           "Statistics value (%s) must be a non null positive integer value",
           _p);
   if ( lib_is_numeric(_p) )
   {
        return(-1);
   }
   _statvalue = atoi(_p);
   if ( _statvalue <= 0 )
   {
        return(-1);
   }
   c->statistics = ENABLE;
   c->statisticsperiod = _statvalue;
   return(0);
}

/* void encode_mission_file()
 * Encode a mission_file structure into a encoded_mission_file structure
 *
 * Argument : pointer to mission_file structure
 *            pointer to encoded_mission_file structure
 *
 */
void encode_mission_file(cp,ecp)
struct mission_file          *cp;
struct encoded_mission_file *ecp;
{
int _i;
int _j;

 if( ecp && cp )
 {
     strncpy(ecp->date,cp->date,DATE_LEN);
     for(_i=0;_i<VERSION_WORDS_COUNT;_i++)
          ecp->version[_i]= cp->version[_i];
     for(_i=0;_i<MEMBER_COUNT;_i++)
     {
         strncpy(ecp->memberlist[_i].name,
                 cp->memberlist[_i].name,
                 MEMBER_LEN);
/*
         lib_i_to_char((WDX_INT *)&cp->memberlist[_i].type,
                       (WDX_BYTE *)&ecp->memberlist[_i].type);
         lib_i_to_char((WDX_INT *)&cp->memberlist[_i].priority,
                       (WDX_BYTE *)&ecp->memberlist[_i].priority);
         lib_i_to_char((WDX_INT *)&cp->memberlist[_i].broadcast,
                       (WDX_BYTE *)&ecp->memberlist[_i].broadcast);
         lib_i_to_char((WDX_INT *)&cp->memberlist[_i].validity,
                       (WDX_BYTE *)&ecp->memberlist[_i].validity);
         lib_i_to_char((WDX_INT *)&cp->memberlist[_i].redundancy,
                       (WDX_BYTE *)&ecp->memberlist[_i].redundancy);
*/
         lib_i_to_char(&cp->memberlist[_i].type,
                       (WDX_BYTE *)&ecp->memberlist[_i].type);
         lib_i_to_char(&cp->memberlist[_i].priority,
                       (WDX_BYTE *)&ecp->memberlist[_i].priority);
         lib_i_to_char(&cp->memberlist[_i].broadcast,
                       (WDX_BYTE *)&ecp->memberlist[_i].broadcast);
         lib_i_to_char(&cp->memberlist[_i].validity,
                       (WDX_BYTE *)&ecp->memberlist[_i].validity);
         lib_i_to_char(&cp->memberlist[_i].redundancy,
                       (WDX_BYTE *)&ecp->memberlist[_i].redundancy);
         for(_j=0;_j<NETWORKS;_j++)
             strncpy(ecp->memberlist[_i].netname[_j],
                     cp->memberlist[_i].netname[_j],
		     IFNAMSIZ);
     }
 }
}

/* int conf_checksum()
 * Calculate a checksum on conf_file data base.
 * The mission_file database must have a number of lines multiple of size of int.
 * The checksum is calculated using columns to avoid lines compensations
 *
 * Argument : pt to mission_file
 *            pt to checksum
 *
 * Return   : 0 ok
 *            -1 error
 */
int conf_checksum(cfpt,cspt)
struct mission_file *cfpt;
unsigned int       *cspt;
{
unsigned char *_pt;
unsigned int _cs0;
unsigned int _cs1;
unsigned int _cs2;
unsigned int _ci;
unsigned int _c;
int _i,_j,_k;
struct encoded_mission_file _ecf;

   if ( MEMBER_COUNT % (sizeof(int)) != 0 )
   {
        PERROR_B
        "MEMBER_COUNT bad compilation value"
        PERROR_E;
        return(-1);
   }
   if ( DATE_LEN % (sizeof(int)) != 0 )
   {
        PERROR_B
        "DATE_LEN bad compilation value"
        PERROR_E;
        return(-1);
   }

/* build an operating system independant buffer */
   memset(&_ecf,'\0',sizeof(struct encoded_mission_file));
   encode_mission_file(cfpt,&_ecf);

/* calculate date checksum */
   _cs0 = 0;
   for(_j=0;_j<DATE_LEN;_j+=INT_SIZE)
   {
       _ci = 0;
       for(_i=0;_i<INT_SIZE;_i++)
       {
            _pt = (unsigned char *)(&_ecf.date[_i+_j]);
            _c  = (unsigned int) *(_pt);
            _c <<= (INT_IN_BITS - BYTE_IN_BITS);
            _c >>= (INT_IN_BITS - BYTE_IN_BITS);
            _ci <<= BYTE_IN_BITS;
            _ci = _ci | _c;
       }
       _cs0  += _ci;
   }

/* calculate version checksum */
   _cs1 = 0;
   _ci = 0;
   for(_i=0;_i<VERSION_WORDS_COUNT;_i++)
   {
       _c  = (unsigned int) (_ecf.version[_i]);
       _c <<= (INT_IN_BITS - BYTE_IN_BITS);
       _c >>= (INT_IN_BITS - BYTE_IN_BITS);
       _ci <<= (BYTE_IN_BITS);
       _ci = _ci | _c;
   }
   _cs1 = _ci;

/* calculate checksum on member table : each 4 lines have a pseudo height of 'j'*/
   _cs2 = 0;
   for(_k=0;_k<sizeof(struct encoded_member_def);_k++)
   {
        for(_j=0;_j<MEMBER_COUNT;_j+=INT_SIZE)
        {
                _ci = 0;
                for(_i=0;_i<INT_SIZE;_i++)
                {
                    _pt = (unsigned char *)(&_ecf.memberlist[_i+_j]);
                    _c  = (unsigned int) *(_pt + _k);
                    _c <<= (INT_IN_BITS - BYTE_IN_BITS);
                    _c >>= (INT_IN_BITS - BYTE_IN_BITS);
                    _ci <<= BYTE_IN_BITS;
                    _ci = _ci | _c;
                    _ci += _j;
                }
                _cs2 += _ci;
         }
   }
/*
sprintf(messagestring,"Checksum value : %u",_cs0 + _cs1 + _cs2);
PINFOR_B
messagestring
PINFOR_E;
*/
   *cspt =  _cs0 + _cs1 + _cs2;

   return(0);
}

/* int unique_key()
 * algorithm to determine a unique key according to mission name :
 *        MISSION_LEN is defined as 16
 *        MISSION_NAME is restricted to MISSION_LEN characters.
 *        MISSION_NAME = C0 C1 C2 ..... Cn ( Ci is an ASCII character )
 *        KEY = sigma( (1+(decimal value)Ci)*2^(7+i) ) ( i from 0 to
 *        MISSION_LEN -1 )
 *        KEY values in the range 8388480 to 1065336960
 *
 * Return : key value
 */
int unique_key(mission)
char *mission;
{
char base[MISSION_LEN];
int i;
int key;

   memset(base,'\0',sizeof(base));
   strncpy(base,mission,MISSION_LEN);
   key = 0;
   for(i=0;i<MISSION_LEN;i++)
       key += (1+(int)base[i]) * (1<<(7+i));
   return(key);
}

/*  clear_leading_char()
 * Erase all leading non printable char into a string
 *
 * Argument : pt to string
 */
void clear_leading_char(char *c)
{
char *_p;

   for(_p=c;*_p!='\0';_p++)
   {
       if ( isprint(*_p) == 0 || *_p ==' ' )
          *_p = '\0';
   }
}

/* int get_version()
 * Analyze the version line field
 *
 * Argument : pt to version field
 *	      pt to high result
 *	      pt to low result
 *
 * Return : 0 ok / -1 error
 */
int get_version(s,h,l)
char                *s;	/* line read without keyword */
unsigned char       *h; /* high value to fill in */
unsigned char 	    *l; /* low value to fill in */
{
int  _h;
int  _l;
int  _words;

    _words = sscanf(s,"%d.%d",&_h,&_l);
    if ( _words != VERSION_WORDS_COUNT ||
	     ( _l > VERSION_LOW_MAX ) ||
	     ( _h < VERSION_HIGH_MIN || _h > VERSION_HIGH_MAX ) ||
	     ( _l < VERSION_LOW_MIN && _h == VERSION_HIGH_MIN ) )
    {
             sprintf(messagestring,"Invalid version format (%s)",s);
             return(-1);
    }
    *h = (unsigned char ) _h;
    *l = (unsigned char ) _l;
    return(0);
}

/* int get_date()
 * Analyze one date line
 *
 * Argument : pt to date field
 *	      pt to date result
 *
 * Return : 0 ok / -1 error
 */
int get_date(s,r)
char                *s;	/* date read without keyword */
char                *r; /* date result pt */
{
   strncpy(r,s,DATE_LEN);
   if ( strlen(r) != strlen(s) )
   {
        sprintf(messagestring,"Invalid date format (%s)",s);
        return(-1);
   }
   return(0);
}
