/*
 *   WDX: WitchDoctor-WDX, versatile distributed lock manager
 *   Copyright (C) 2003 Dominique Chabord,
 *   Please refer to credits section below
 *
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ************************************************
 *
 *   History and credits:
 *   First version of WitchDoctor-WDX was based on:
 *
 *   DECwatchdog/Autopilot, watchdog and lock manager
 *   Copyright (C) 2003 HP
 *
 *   Developed by
 *     Computer Special Systems, Annecy France
 *     A Division of Digital Equipment Corporation
 *   and
 *     Custom-Systems, Annecy France
 *     An engineering center of Compaq
 *
 *   Provided under GNU-GPL terms by HP, 30 June 2003
 *   GNU-GPL replaces previous license statements
 *
 *   Rebranded, ported and modified for Shaman-X project, 15 July 2003
 *   by Dominique Chabord
 *   and maintained by other contributors as mentionned in the code
 *
 *   ************************************************
 *
 *   WDX is promoted by Shaman-X project
 *
 *   Check for new versions and documentation at
 *   http://www.shaman-x.org
 *
 *   ************************************************
 *
 *   Most of WDX code is designed to run unchanged on
 *   different operating systems, including at V1.0 time:
 *   Gnu-Linux, WindowsNT, DigitalUNIX and OpenVMS.
 *   Operating system specifics are identified by dedicated files.
 *
 *   Evolutions
 *   ----------
 *   July-2003     Dominique Chabord    WitchDoctor-WDX V1.0
 *                 Creation
 *
 *   ************************************************
 *
 *
 *	wdx_library.c
 *
 *	Module including general routines used by WDX
 *
 *	Public routines :
 *
 *      void lib_char_to_i()
 *      void lib_i_to_char()
 *      void lib_encode_message()
 *      void lib_decode_message()
 *      void lib_forward_operation()
 *      void lib_forward_command()
 *      void lib_forward_announce()
 *      unsigned int lib_checksum()
 *      int  lib_redirect_output()
 *      void lib_output_message()
 *      char *lib_generation_to_s()
 *      int lib_index_of_resource()
 *      enum resource_status lib_status_of_resource()
 *      int lib_requester_of_resource();
 *	int lib_is_member_valid();
 *      int lib_create_process_info()
 *      int lib_delete_process_info()
 *      int lib_set_process_info()
 *      int lib_get_process_info()
 *      int lib_set_snapshot()
 *      int lib_get_snapshot()
 *      int lib_process_init()
 *	int lib_build_notification()
 *
 *	Local routines :
 *
 *	void encode_header()
 *	void decode_header()
 *	void encode_watchdog()
 *	void decode_watchdog()
 *	void encode_resource()
 *	void decode_resource()
 *	void encode_heartbeat()
 *	void decode_heartbeat()
 *	void encode_command()
 *	void decode_command()
 *	void encode_operation()
 *	void decode_operation()
 *	void update_group()
 *	void process_communication_exit()
 *	void process_info_exit()
 *	int get_the_index()
 *	void build_header()
 *	void build_watchdog()
 ************************************************************************/
/*
 * Modification history:
 *
 *  22-Mar-94 -- Cornu
 *      Creation
 */

#undef WDX_LIBRARY_MODULE
#define WDX_LIBRARY_MODULE

#include	"wdx.h"

/*
 * ERROR management definition and declarations
 */
#define MODE 0777

static enum flag filelog = ENABLE;
static enum flag redirflag = DISABLE;

/*
 * Variables used by communication protocol : set up by set_process_info() and
 * get_process_info()
 */
static char             wdxversion;
static unsigned int  configversion;
static int              missionkey;

/*
 * PRINT management definition
 */
char month[12][4] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
char generation[STRING_LEN];

/*
 * Local routines prototypes
 */

/*
 * LOCAL ROUTINES
 */
void encode_header(struct header          *hp,struct encoded_header *ehp);
void decode_header(struct encoded_header *ehp,struct header          *hp);
void encode_watchdog(struct watchdog      *wp,struct encoded_watchdog *ewp);
void decode_watchdog(struct encoded_watchdog *ewp,struct watchdog *wp);
void encode_resource(struct resource *rp,struct encoded_resource *erp);
void decode_resource(struct encoded_resource *erp,struct resource *rp);
void encode_heartbeat(struct heartbeat *hp,struct encoded_heartbeat *ehp);
void decode_heartbeat(struct encoded_heartbeat *ehp,struct heartbeat *hp);
void encode_command(struct command *cp, struct encoded_command *ecp);
void decode_command(struct encoded_command *ecp, struct command *cp);
void encode_operation(struct operation *cp, struct encoded_operation *ecp);
void decode_operation(struct encoded_operation *ecp, struct operation *cp);
void build_header(struct header *p,enum msg_type t,int s,unsigned int n);
void build_watchdog(struct watchdog *p,struct member *m);
void update_group(struct group *n, struct group *o);
void process_communication_exit(void);
void process_info_exit(void);
int get_the_index(struct group *g,char *hname);



/*
 * PUBLIC ROUTINES : prefixe is "lib"
 */


/* void lib_char_to_i()
 * Convert a integer coded as a 4 bytes into an integer
 *
 * Argument : pointer to 4 WDX_BYTES
 *          : pointer to WDX_INT structure
 *
 * Return   : none
 */

/*
  Modified by DM 10/10/03
  To cater for the MAC where intergers are not stored in reverse byte order
  4 characters are converted to an unsigned int.
  All calls to this routine have been modified to change the parameter type
*/
void lib_char_to_i (WDX_BYTE *ch, unsigned int *i)
{
int j;

	*i = 0;

	for (j = 0; j < 32; j = j + 8, ch++)
		*i = *i + (((unsigned char)(*ch)) << j);

}

/*
 void lib_char_to_i(ch,i)
WDX_BYTE *ch;
WDX_INT  *i;
{
  i->c0 = *ch++;
  i->c1 = *ch++;
  i->c2 = *ch++;
  i->c3 = *ch;
}
*/


/* void lib_i_to_char()
 * Code a 32 bit integer into on 4 BYTES
 *
 * Argument : pointer to WDX_BYTE of 4 bytes
 *            pointer to WDX_INT structure
 *
 * Return   : none
 */
/*
  Modified by DM 10/10/03
  To cater for the MAC where intergers are not stored in reverse byte order
  Unsigned int is converted to 4 characters.
  All calls to this routine have been modified to change the parameter type
*/
void lib_i_to_char (unsigned int *i, WDX_BYTE *ch)
{
int j;

	for (j = 0; j < 32; j = j + 8)
		*ch++ = (unsigned char)((*i >> j) & 0xff);
}

/*

void lib_i_to_char(i,ch)
WDX_INT  *i;
WDX_BYTE *ch;
{

  *ch++ = i->c0;
  *ch++ = i->c1;
  *ch++ = i->c2;
  *ch = i->c3;
}
*/


/* void lib_encode_message()
 * Encode a message structure into a encoded message structure
 *
 * Argument : pt to message structure
 *            pt to encoded_message structure
 */
void lib_encode_message(mp,emp)
struct message *mp;
struct encoded_message *emp;
{
int i;

   if ( mp && emp )
        {
        encode_header(&(mp->head),&(emp->head));
        switch(mp->head.datatype)
             {
             case HEARTBEAT :
                  encode_heartbeat(&(mp->m.h),&(emp->m.h));
                  break;
             case COMMAND :
                  encode_command((struct command *)&(mp->m.c),
                                 (struct encoded_command *)&(emp->m.c));
                  break;
             case OPERATION :
                  encode_operation((struct operation *)&(mp->m.o),
                                   (struct encoded_operation *)&(emp->m.o));
                  break;
             default :
                  break;
             }
        for(i=0;i<PAD_LEN;i++)
             emp->pad[i] = mp->pad[i];
        }
}

/* void lib_decode_message()
 * Decode a encoded_message structure into a message structure
 *
 * Argument : pt to encoded_message structure
 *            pt to message structure
 */
void lib_decode_message(emp,mp)
struct encoded_message *emp;
struct message *mp;
{
int i;

   if ( mp && emp )
        {
        decode_header(&(emp->head),&(mp->head));
        switch(mp->head.datatype)
             {
             case HEARTBEAT :
                  decode_heartbeat(&(emp->m.h),&(mp->m.h));
                  break;
             case COMMAND :
                  decode_command((struct encoded_command *)&(emp->m.c),
                                 (struct command *)&(mp->m.c));
                  break;
             case OPERATION :
                  decode_operation((struct encoded_operation *)&(emp->m.o),
                                   (struct operation *)&(mp->m.o));
                  break;
             default :
                  break;
             }
        for(i=0;i<PAD_LEN;i++)
             mp->pad[i] = emp->pad[i];
        }
}

/* void lib_forward_operation()
 * Build then forward an operation message to the local WDX process
 *
 * Argument : socket descriptor
 *            operation type
 *	      operation identifier
 *            operation data
 *	      operation argument
 *
 * Return   :
 */
void lib_forward_operation(descriptor,t,i,d,a)
int       descriptor;	    /* descriptor */
int                t;       /* operation type */
int                i;	    /* operation identifier */
char              *d;       /* data pointer */
int                a;	    /* operation argument */
{
struct message mess;
struct encoded_message emess;
struct socket_output output;
unsigned int lc;		/* local checksum value */
struct operation *p;  /* command pointer */

   p = &(mess.m.o);
   if(p)
   {
    memset(&mess,'\0',sizeof(struct message));
    memset(&emess,'\0',sizeof(struct encoded_message));
    /* build header */
    build_header(&(mess.head),OPERATION,sizeof(struct operation),
                 DATA_NUMBER_OPERATION);

    /* create operation message */
    p->index = localindex;
    p->pid = getpid();
    p->type = t;
    p->identifier = i;
    if (d)
       strncpy(p->operationdata,d,OPERATION_LEN);
    p->operationargument = a;

    /* build encoded message */
    lib_encode_message(&mess,&emess);

    /* add checksum to message : checksum is encoded */
    lc = lib_checksum((unsigned char *)&(emess.CHECK_MESSAGE),SIZE_MESSAGE);
//    lib_i_to_char((WDX_INT *)&lc,(WDX_BYTE *)&emess.checksum);
    lib_i_to_char(&lc,(WDX_BYTE *)&emess.checksum);
    output.descriptor = descriptor;
    output.buffer = (char *) &emess;
    output.size = sizeof(struct encoded_message);
#ifndef DEBUG
    com_send_socket_message(&output);
#endif
   }
}

/* void lib_forward_command()
 * Build then forward an command message to all the WDX processes
 *
 * Argument : socket descriptor
 *            command type
 *	      command data pointer
 *	      data number
 *
 * Return   :
 */
void lib_forward_command(descriptor,t,d,n)
int       descriptor;	    /* descriptor */
enum command_type  t;       /* command type */
char              *d;       /* data pointer */
unsigned int       n;	    /* data number */
{
struct message mess;
struct encoded_message emess;
struct socket_output output;
unsigned int      lc; /* local checksum value */
struct c_action  *ca; /* c_action pointer */
struct c_request *cr; /* c_request pointer */
struct c_answer  *cw; /* c_answer pointer */
struct command    *p; /* pt to command data structure in message */

   p = &(mess.m.c);
   if(p)
   {
    memset(&mess,'\0',sizeof(struct message));
    memset(&emess,'\0',sizeof(struct encoded_message));
    /* build header */
    build_header(&(mess.head),COMMAND,sizeof(struct command),n);

    /* create operation message */
    p->index = localindex;
    p->type = t;
    switch(t)
    {
    case TREQUEST :
	cr = (struct c_request *) &(p->value1);
        memcpy(cr,d,sizeof(struct c_request));
	break;
    case TACTION :
	ca = (struct c_action *) &(p->value1);
        memcpy(ca,d,sizeof(struct c_action));
	break;
    case TANSWER :
	cw = (struct c_answer *) &(p->value1);
        memcpy(cw,d,sizeof(struct c_answer));
	break;
    }

    /* build encoded message */
    lib_encode_message(&mess,&emess);

    /* add checksum to message : checksum is encoded */
    lc = lib_checksum((unsigned char *)&(emess.CHECK_MESSAGE),SIZE_MESSAGE);
//    lib_i_to_char((WDX_INT *)&lc,(WDX_BYTE *)&emess.checksum);
    lib_i_to_char(&lc,(WDX_BYTE *)&emess.checksum);
    output.descriptor = descriptor;
    output.buffer = (char *) &emess;
    output.size = sizeof(struct encoded_message);
#ifndef DEBUG
    com_send_socket_message(&output);
#endif
   }
}

/* void lib_forward_announce()
 * Build a HEARTBEAT message
 *
 * Argument : Pt to message structure
 *	      pt to local member structure
 *            data number value
 */
void lib_forward_announce(struct message *mess,struct member *localmemberp,unsigned int dtn)
{
int lc;
struct socket_output        output;    /* output socket data structure */
struct encoded_message out_message;    /* outgoing watchdog message */

    /* build message */
    build_watchdog(&(mess->m.h.wat),localmemberp);

    build_header(&(mess->head),HEARTBEAT,sizeof(struct heartbeat),dtn);

    /* build encoded message */
    lib_encode_message(mess,&out_message);

    /* add checksum to message : checksum is also encoded */
    lc = lib_checksum((unsigned char *)&(out_message.CHECK_MESSAGE),
                      SIZE_MESSAGE);
//    lib_i_to_char((WDX_INT *)&lc,(WDX_BYTE *)&out_message.checksum);
    lib_i_to_char(&lc,(WDX_BYTE *)&out_message.checksum);
    output.descriptor = socketdescriptor;
    output.buffer = (char *) &out_message;
    output.size = sizeof(struct encoded_message);

#ifndef DEBUG
    com_send_socket_message(&output);
#endif
}

/* unsigned int lib_checksum()
 *
 * A checksum calculated on an array of SIZE integer is returned :
 *
 * Argument : pointer to char array
 *            size
 *
 * Return   : checksum
 */
unsigned int lib_checksum(l,size)
unsigned char *l;
int size;
{
unsigned int cs;
unsigned int c;
unsigned int c1;
int i,j,k;

    cs = 0;
    c = 0;
    i = 0;
    j = 0;
    k = 0;

    for (k = 0; k < (int)(size-INT_SIZE); k+=INT_SIZE)
    {
        c1 = 0;
        for(j = 0; j < INT_SIZE; j++,i++)
        {
                c = *(l+i);
                c <<= (INT_IN_BITS - BYTE_IN_BITS);
                c >>= (INT_IN_BITS - BYTE_IN_BITS);
                c1 <<= BYTE_IN_BITS;
                c1 = c1 | c;
        }
        cs += c1;
    }
    return(cs);
}


/* int lib_redirect_output()
 * This subroutine redirect the standard error stderr to the file given as
 * argument.
 * A file having the same name is renamed or reopen.
 *
 * Arguments : log flag
 *	       log file name
 *
 * Returns:    0 if OK
 *            -1 if error
 */
int lib_redirect_output(flg, log)
enum flag flg;
char     *log;
{
   if ( flg == DISABLE )
   {
	filelog = DISABLE;
	return(0);
   }

/* rename log file if already exist */
   fclose(stderr);
   spc_rename(log);

/* open log file */
   if ( freopen(log,"w+",stderr) == NULL )
   {
        PERROR_B
        "Can't redirect stderr to log file"
        PERROR_E;
        return(-1);
   }
   redirflag = ENABLE;	/* redirect to file */
   filelog = ENABLE;
   return(0);
}

/* void lib_output_message()
 * print a message to sdterr
 * If one of the arguments include ERROR_TOKEN this token is replaced by
 * the system error message associated with errno.
 *
 * Argument : variable list of :
 *		- one integer
 *		- string pointer
 *            list terminated with (char *) NULL
 *
 * Return   : none
 */
void lib_output_message(int first,...)
{
   va_list ap;
   char *p;
   char *pt;
   char *args[MAX_ARG];
   int  argno;
   char local_string1[STRING_LEN];
   char local_string2[STRING_LEN];
   time_t  gen;
   int fdate;

	va_start(ap,first);
	if ( filelog == ENABLE)
	{
		switch(first)
		{
		case WDX_MESS_E : /* error */

			if ((redirflag == ENABLE))
				p = ERROR_MESSAGE;
			else
				p = ERROR_MESSAGE_SPECIFIC;

			fdate = 1;
            break;

		case WDX_MESS_I : /* information */
			if ((redirflag == ENABLE))
				p = INFOR_MESSAGE;
			else
				p = INFOR_MESSAGE_SPECIFIC;

			fdate = 1;
			break;

		case WDX_MESS_W : /* warning */
			if ((redirflag == ENABLE))
				p = WARNING_MESSAGE;
			else
				p = WARNING_MESSAGE_SPECIFIC;

			fdate = 1;
			break;

		case WDX_MESS_C : /* complement */
			p = COMPLEMENT_MESSAGE;
			fdate = 0;
			break;
		}

		if ( fdate && (redirflag == ENABLE))
		{
			gen = lib_build_generation();
			fprintf(stderr,"%s%s",p,lib_generation_to_s(&gen));
		}
		else
			fprintf(stderr,"%s",p);

		args[0] = (char *) 0;
		argno = 0;

		while (args[argno++] = va_arg(ap, char *))
		{
			if ( fdate && (redirflag == ENABLE) )
			{
				if((argno - 1) >= 0)
					fprintf(stderr,", ");
			}
			else
				if((argno - 1) >= 1)
					fprintf(stderr,", ");

			if ( (pt = strstr(args[argno - 1], ERRNO_TOKEN)) != NULL )
			{
/*
                   *(++pt) = 's';
                   sprintf(local_string1,"\n%s",strerror(errno));
                   sprintf(local_string2,args[argno-1],local_string1);
                   fprintf(stderr,"%s",local_string2);
                   fprintf(stderr," %d ",errno);
*/
/*++ 25-May-94 BC : errno is not always in the UNIX range */
				*(++pt) = 's';
				sprintf(local_string1,"(%d)",errno);
				sprintf(local_string2,args[argno-1],local_string1);
				fprintf(stderr,"%s",local_string2);
/*-- 25-MAY-94 BC */
			}
			else
			fprintf(stderr,"%s",args[argno - 1]);

		}

		fprintf(stderr,"\n");
		fflush(stderr);
	}
}

/* char * lib_generation_to_s()
 * Converts generation time to string
 *
 */
char *lib_generation_to_s(time_t *g)
{
struct tm *tm;
int year;

  tm = spc_localtime(g);
   if ( tm->tm_year < 90 )
        year = tm->tm_year + 2000;
   else year = tm->tm_year + 1900;
   sprintf(generation,"%02.2d-%s-%d %02.2d:%02.2d:%02.2d",
           tm->tm_mday,month[tm->tm_mon],year,
           tm->tm_hour,tm->tm_min,tm->tm_sec);
   return(generation);
}

/* int lib_index_of_resource()
 * Returns index in resource status table of a resource pointed by res
 * If resource not found -1 is returned
 *
 * Argument : pointer to resource area
 *            pointer to resource name
 *
 * Return   : index
 *            NOT_AN_INDEX if not found
 */
int lib_index_of_resource(rsd,res)
struct resource_def *rsd;
char *res;
{
int i;
int ret;

    ret = NOT_AN_INDEX;
    if (res)
        {
	for (i=0; i < RESOURCE_COUNT; i++)
	    if  (strcmp(rsd[i].name,res)==0)
                {
		ret = i;
		break;		/* only the first hit is of interest	*/
                }
        }
    return (ret);
}

/* enum resource_status lib_status_of_resource()
 * Determines the status of the resource
 *
 * Argument : pointer to resource area
 *            pointer to resource name
 *
 * Return   : resource status
 *
 */
enum resource_status lib_status_of_resource(rsd,p)
struct resource_def *rsd;
char *p;
{
int i;

    if ( (i= lib_index_of_resource(rsd,p)) != NOT_AN_INDEX )
         return((rsd+i)->status);
    else return(NOTKNOWN);
}

/* int lib_requester_of_resource()
 * Returns the "initiator" index of a resource
 *
 * Argument : pointer to resource area
 *            pointer to resource name
 *
 * Return   : "initiator" index or NOT_AN_OWNER
 *
 */
int lib_requester_of_resource(rsd,p)
struct resource_def *rsd;
char *p;
{
int i;

   if ( (i= lib_index_of_resource(rsd,p)) != NOT_AN_INDEX )
        return((rsd+i)->initiator);
   else return(NOT_AN_OWNER);
}

/* int lib_is_member_valid()
 * Check the validity of the network interfaces of a member
 * refered by its index
 *
 * Argument : index
 *
 * Return   : 0 not valid / 1 valid
 */
int lib_is_member_valid(mfp,ind)
struct member *mfp;
int ind;
{
int i;
int ret;
struct member *m;

   ret = 0;
   m = mfp + ind;
   for(i=0;i<NETWORKS;i++)
   {
        ret = ret | ( m->networks[i].status == ON );
   }
   return(ret);
}

/* lib_create_process_info()
 * Create the mission_info database : empty file
 *
 * Argument : pt to mission name
 *
 * Return   : 0 on success or -1 on error
 *
 */
int lib_create_process_info(m)
char *m;		/* pt to mission name */
{
FILE *_fdes;
char _file[STRING_LEN];
struct member_base _b;

  sprintf(_file,DATA_BASE,wdxdirectory,m);
  /* Is file already existing */
  if ( lib_get_process_info(m,&_b) == 0 )
  {
       if ( spc_process_running(_b.pid) == 0 )
            return(-1);
  }
  /* Create empty file */
  if ((_fdes = fopen(_file,"w+")) == NULL )
       return(-1);
  fclose(_fdes);
  return(0);
}

/* int lib_delete_process_info()
 * Delete the mission_info database.
 *
 * Argument : pt to mission name
 *
 * Return   : 0 on success or -1 on error
 *
 */
int lib_delete_process_info(m)
char *m;		/* pt to mission name */
{
char _file[STRING_LEN];
int  _ret;
  sprintf(_file,DATA_BASE,wdxdirectory,m);

  _ret = spc_unlink(_file);
  if ( _ret == -1 )
  {
    PERROR_B
    "Can't delete wdx process mission database"
    PERROR_E;
  }
  return(0);
}

/* int lib_set_process_info()
 * Set the mission_info database.
 *
 * Argument : pt to mission name
 *	      pt to member_base structure
 *
 * Return   : 0 on success or -1 on error
 *
 */
int lib_set_process_info(m,p)
char *m;		/* pt to mission name */
struct member_base *p;  /* pt to member */
{
FILE *_fdes;
char _file[STRING_LEN];
int _nb;

/* Set bata_base info */
  sprintf(_file,DATA_BASE,wdxdirectory,m);
  if ((_fdes = fopen(_file,"w")) == NULL )
       return(-1);
  _nb = fwrite(p,sizeof(struct member_base),1,_fdes);
  fclose(_fdes);
  if ( _nb != 1 )
       return(-1);
  wdxversion = p->version;
  configversion = p->checksum;
  missionkey = p->key;
  localindex = p->localindex;
  return(0);
}

/* int lib_get_process_info()
 * Read the mission_info database. Set the global variable wdxversion to the
 * value read.
 *
 * Argument : pt to mission name
 *	      pt to member_base structure
 *
 * Return   : 0 on success or -1 on error
 *
 */
int lib_get_process_info(m,p)
char *m;		/* pt to mission name */
struct member_base *p;  /* pt to member_base */
{
FILE *_fdes;
char _file[STRING_LEN];
int _nb;

  sprintf(_file,DATA_BASE,wdxdirectory,m);
  if ((_fdes = fopen(_file,"r")) == NULL )
       return(-1);
  _nb = fread(p,sizeof(struct member_base),1,_fdes);
  fclose(_fdes);
  if ( _nb != 1 )
       return(-1);
  wdxversion = p->version;
  configversion = p->checksum;
  localindex = p->localindex;
  missionkey = p->key;
  return(0);
}

/* int lib_set_snapshot()
 * Create a snapshot :
 *	- file creation
 *	- group structure saved in file
 *
 * Argument : pt to snapshot name
 *	      pt to group structure
 *
 * Return   : 0 on success or -1 on error
 *
 */
int lib_set_snapshot(s,g)
char         *s; /* pt to snapshot name */
struct group *g; /* pt to group */
{
FILE *_fdes;
char _file[STRING_LEN];
int _nb;

  if(!s || !g)
     return(-1);
  sprintf(_file,SNAPSHOT_FILE,wdxdirectory,g->mission,s);
  /* delete old snapshot files */
  spc_unlink(_file);
  /* Create file */
  if ((_fdes = fopen(_file,"w+")) == NULL )
  {
       PERROR_B
       "Can't create snapshot"
       PERROR_E;
       return(-1);
  }
  _nb = fwrite(g,sizeof(struct group),1,_fdes);
  fclose(_fdes);
  if(_nb != 1)
  {
       PERROR_B
       "Snapshot write error"
       PERROR_E;
       return(-1);
  }
  return(0);
}

/* lib_get_snapshot()
 * Read the snapshot file. If the snapshot read was not for the specified
 * mission, then return an error
 *
 * Argument : pt to snapshot name
 *	      pt to mission name
 *	      pt to group structure to fill in
 *
 * Return   : 0 on success or -1 on error
 *
 */
int lib_get_snapshot(s,m,g)
char         *s; /* pt to snapshot name */
char         *m; /* pt to mission name */
struct group *g; /* pt to group to fill in */
{
FILE *_fdes;
char _file[STRING_LEN];
int _nb;

  if(!s || !g || !m)
      return(-1);

  sprintf(_file,SNAPSHOT_FILE,wdxdirectory,m,s);
  if ((_fdes = fopen(_file,"r")) == NULL )
  {
       PERROR_B
       "Can't open snapshot"
       PERROR_E;
       return(-1);
  }
  _nb = fread(g,sizeof(struct group),1,_fdes);
  fclose(_fdes);
  if ( _nb != 1 )
  {
	PERROR_B
	"Snapshot read error"
	PERROR_E;
       return(-1);
  }
  /* check for mission name */
  if(strcmp(g->mission,m)!=0)
  {
       PERROR_B
       "Snapshot and mission do not match"
       PERROR_E;
       return(-1);
  }
  return(0);
}

#ifndef WDX_BUILD_API
/* int lib_process_init()
 * Initialize or re-initialize the process.
 * Several steps are performed at initialisation according to init_mode value :
 *  COLD_INIT_PROCESS :
 *	- set up wdx path
 *	- create process data base file
 *	- cfg_config_init()
 *	- init local index
 *	- set process data
 *	- init communication
 *	- set process data in process data file
 *	- init linked modules with COLD_INIT argument
 *      Arguments are : s points to mission name
 *			g1 points to data base
 *		        g2 not used
 *			m points to message structure
 *  COLD_REINIT_PROCESS :
 *	- init linked modules with COLD_REINIT argument
 *      Arguments are : s not used
 *			g1 points to data base
 *		        g2 not used
 *			m points to message structure
 *  WARM_REINIT_PROCESS :
 *	- get snapshot then update group
 *	- init local index
 *	- set process data
 *	- re-init communication
 *	- set process data in process data file
 *	- init linked modules with COLD_REINIT or WARM_INIT
 *      Arguments are : s points to snapshot name
 *			g1 points to final data base
 *		        g2 points to read data base
 *			m points to message structure
 *
 * Argument : COLD_INIT_PROCESS or COLD_REINIT_PROCESS or WARM_REINIT_PROCESS
 *	      string pt : mission name or snapshot name
 *	      group data base pt
 *	      tmp group data base pt
 *	      message structure pt
 */
int lib_process_init(c,s,g1,g2,m)
enum init_mode c;  /* init mode */
char          *s;  /* pt to string */
struct group *g1;  /* main group data structure : will contain the data base */
struct group *g2;  /* working group */
struct message *m; /* message pt */
{
struct member			   *pt; /* pointer to local member structure */
struct member_base                data; /* member data structures */
int				     i; /* index */
struct function_table             *tpt; /* pt to initable function table */
enum startup_mode                start; /* start type */
int                                ret;	/* returned code */
int                               step; /* step */
struct group			     g; /* working group data structure */

#define INIT_PATH 0
#define INIT_PROCESS_FILE 5
#define INIT_STOP 10
#define INIT_CONFIG 1
#define INIT_LOCAL_INDEX 2
#define INIT_PROCESS_DATA 3
#define INIT_COMMUNICATION 4
#define INIT_PROCESS_INFO 6
#define INIT_MODULE 7

//printf("in lib_process_init(\n");

    ret = 0;		/* return success */
    start = COLD_INIT;  /* cold init : the default */
    step = INIT_PATH;
    if ( c == COLD_REINIT_PROCESS )
    {
	step = INIT_MODULE;
	start = COLD_REINIT;
    }
    memset(&data,'\0',sizeof(struct member_base));

    while(step != INIT_STOP)
    {
//printf("while\n");
	switch(step)
	{
	case INIT_PATH : /* initialize wdx path */
	    if ( c == COLD_INIT_PROCESS )
                 if ( spc_path_init(&wdxdirectory) != 0 )
	    {
		 PERROR_B
		 "Wdx path not defined"
		 PERROR_E;
		 ret = -1;
		 step = INIT_STOP;
		 break;
	    }
	    step = INIT_PROCESS_FILE;
	    break;
	case INIT_PROCESS_FILE : /* first step : create process info file */
	    if ( c == COLD_INIT_PROCESS )
	    {
		if ( (ret = lib_create_process_info(s)) == -1 )
		{
		    PERROR_B
		    "Wdx process already started"
		    PERROR_E;
		    ret = -1;
		    step = INIT_STOP;
		    break;
		}
	    atexit(process_info_exit);
	    }
	    step = INIT_CONFIG;
	    break;
	case INIT_CONFIG : /* read config file or snapshot */
	    if ( c == COLD_INIT_PROCESS )
	    {
                 if ( cfg_config_init(s,g1) == -1 )
		 {
		    ret  = -1;
		    step = INIT_STOP;
		    break;
		 }
	    }
	    else
	    {
                 memset(&g,'\0',sizeof(struct group));
                 if ( lib_get_snapshot(s,g1->mission,&g) == 0 )
		 {
                      update_group(g2,&g); /* update_group */
	              /* copy to real data base */
	              oldversion = g1->configversion;
	              start = WARM_REINIT;  /* warm init */
		 }
		 else start = COLD_REINIT;
		 /* cold and warm startup */
                 memcpy(g1,g2,sizeof(struct group));
	    }
	    step = INIT_LOCAL_INDEX;
	    break;
	case INIT_LOCAL_INDEX : /* init local index data */
            if ((g1->localindex = get_the_index(g1,localname)) == NOT_AN_INDEX )
            {
                sprintf(messagestring,
                "Cannot find member name \"%s\" in mission file",
                 localname);
                PERROR_B
                messagestring
                PERROR_E;
                ret  = -1;
		step = INIT_STOP;
            }
	    else
	    {
                localindex = g1->localindex;
		pt = &(g1->memberlist[localindex]);
		step = INIT_PROCESS_DATA;
	    }
	    break;
	case INIT_PROCESS_DATA : /* set process data base structure */
            strcpy(data.name,pt->name);
            data.redundancy = pt->redundancy;
            data.localindex = localindex;
	    data.type = pt->type;
	    data.pid = getpid();
            for(i=0;i<NETWORKS;i++)
            {
                 strcpy(data.links[i].ifname,pt->networks[i].ifname);
                 data.links[i].use = pt->networks[i].use;
            }
            strcpy(data.udp,g1->mission);
            data.checksum = g1->configversion;
            data.key = g1->missionkey;
            data.version = g1->wdxversion;
	    step = INIT_COMMUNICATION;
	    break;
	case INIT_COMMUNICATION : /* init communication */
	    if ( c == COLD_INIT_PROCESS )
	    {
                 if ( (socketdescriptor = com_open_socket(g1->mission,
		                  pt->redundancy,
			          &(data.links[0]),
				  WDX_PROCESS)) == -1 )

                 {
		    ret = -1;
		    step = INIT_STOP;
		    break;
		 }
		 atexit(process_communication_exit);
	    }
	    else
	    {
	 	 if( com_update_socket(socketdescriptor,
				       pt->redundancy,
				       &(data.links[0])) == -1 )
                 {
	            ret = -1;
		    step = INIT_STOP;
		    break;
		 }
	    }
            localbroadcast = g1->memberlist[localindex].broadcast;
	    step = INIT_PROCESS_INFO;
	    break;
	case INIT_PROCESS_INFO : /* set process data base */
            /* set data base */
            if ( lib_set_process_info(g1->mission,&data) == -1 )
            {
	         PERROR_B
	         "Cannot initialise wdx process mission database "
	         PERROR_E;
	         ret = -1;
		 step = INIT_STOP;
		 break;
            }
        case INIT_MODULE : /* init all the modules */
            tpt = inittable;
            while(tpt->functpt)
            {
                 (*tpt->functpt)(start,g1,m);
                 tpt++;
            }
	    step = INIT_STOP;
	    break;
	default :
	    break;
	}
    }

   return(ret);
}
#endif /* WDX_BUILD_API */

/*
 * LOCAL ROUTINES
 */

/* void encode_header()
 * Encode a header structure into a encoded_header structure
 *
 * Argument : pointer to header structure
 *            pointer to encoded_header structure
 *
 */
void encode_header(hp,ehp)
struct header          *hp;
struct encoded_header *ehp;
{
 if( ehp && hp )
     {
     ehp->wdxversion = hp->wdxversion;
/*
     lib_i_to_char((WDX_INT *)&hp->configversion,
                   (WDX_BYTE *)&ehp->configversion);
     lib_i_to_char((WDX_INT *)&hp->missionkey,
                   (WDX_BYTE *)&ehp->missionkey);
     lib_i_to_char((WDX_INT *)&hp->memberindex,
                   (WDX_BYTE *)&ehp->memberindex);
     lib_i_to_char((WDX_INT *)&hp->memberpid,
                   (WDX_BYTE *)&ehp->memberpid);
     lib_i_to_char((WDX_INT *)&hp->datanumber,
                   (WDX_BYTE *)&ehp->datanumber);
     lib_i_to_char((WDX_INT *)&hp->datatype,
                   (WDX_BYTE *)&ehp->datatype);
     lib_i_to_char((WDX_INT *)&hp->datasize,
                   (WDX_BYTE *)&ehp->datasize);
*/
     lib_i_to_char(&hp->configversion,
                   (WDX_BYTE *)&ehp->configversion);
     lib_i_to_char(&hp->missionkey,
                   (WDX_BYTE *)&ehp->missionkey);
     lib_i_to_char(&hp->memberindex,
                   (WDX_BYTE *)&ehp->memberindex);
     lib_i_to_char(&hp->memberpid,
                   (WDX_BYTE *)&ehp->memberpid);
     lib_i_to_char(&hp->datanumber,
                   (WDX_BYTE *)&ehp->datanumber);
     lib_i_to_char(&hp->datatype,
                   (WDX_BYTE *)&ehp->datatype);
     lib_i_to_char(&hp->datasize,
                   (WDX_BYTE *)&ehp->datasize);
     }
}

/* void decode_header()
 * Decode a encoded_header structure into a header structure
 *
 * Argument : pointer to encoded_header structure
 *            pointer to header structure
 *
 */
void decode_header(ehp,hp)
struct encoded_header *ehp;
struct header          *hp;
{
 if ( ehp && hp )
      {
      hp->wdxversion = ehp->wdxversion;
/*
      lib_char_to_i((WDX_BYTE *)&ehp->configversion,
                    (WDX_INT *)&hp->configversion);
      lib_char_to_i((WDX_BYTE *)&ehp->missionkey,
                    (WDX_INT *)&hp->missionkey);
      lib_char_to_i((WDX_BYTE *)&ehp->memberindex,
                    (WDX_INT *)&hp->memberindex);
      lib_char_to_i((WDX_BYTE *)&ehp->memberpid,
                    (WDX_INT *)&hp->memberpid);
      lib_char_to_i((WDX_BYTE *)&ehp->datanumber,
                    (WDX_INT *)&hp->datanumber);
      lib_char_to_i((WDX_BYTE *)&ehp->datatype,
                    (WDX_INT *)&hp->datatype);
      lib_char_to_i((WDX_BYTE *)&ehp->datasize,
                    (WDX_INT *)&hp->datasize);
*/
      lib_char_to_i((WDX_BYTE *)&ehp->configversion,
                    &hp->configversion);
      lib_char_to_i((WDX_BYTE *)&ehp->missionkey,
                    &hp->missionkey);
      lib_char_to_i((WDX_BYTE *)&ehp->memberindex,
                    &hp->memberindex);
      lib_char_to_i((WDX_BYTE *)&ehp->memberpid,
                    &hp->memberpid);
      lib_char_to_i((WDX_BYTE *)&ehp->datanumber,
                    &hp->datanumber);
      lib_char_to_i((WDX_BYTE *)&ehp->datatype,
                    &hp->datatype);
      lib_char_to_i((WDX_BYTE *)&ehp->datasize,
                    &hp->datasize);
      }
}

/* void encode_watchdog()
 * Encode a watchdog structure into a encoded_watchdog structure
 *
 * Argument :pointer to watchdog structure
 *           pointer to encoded_watchdog structure
 */
void encode_watchdog(wp,ewp)
struct watchdog *wp;
struct encoded_watchdog *ewp;
{
  if ( wp && ewp )
       {
//       lib_i_to_char((WDX_INT *)&wp->index,(WDX_BYTE *)&ewp->index);
//       lib_i_to_char((WDX_INT *)&wp->status,(WDX_BYTE *)&ewp->status);
//       lib_i_to_char((WDX_INT *)&wp->validity,(WDX_BYTE *)&ewp->validity);
       lib_i_to_char(&wp->index,(WDX_BYTE *)&ewp->index);
       lib_i_to_char(&wp->status,(WDX_BYTE *)&ewp->status);
       lib_i_to_char(&wp->validity,(WDX_BYTE *)&ewp->validity);
       }
}

/* void decode_watchdog()
 * Decode a encoded_watchdog structure into a watchdog structure
 *
 * Argument :pointer to edcoded_watchdog structure
 *           pointer to watchdog structure
 */
void decode_watchdog(ewp,wp)
struct encoded_watchdog *ewp;
struct watchdog *wp;
{
  if ( ewp && wp )
       {
/*
       lib_char_to_i((WDX_BYTE *)&ewp->index,(WDX_INT *)&wp->index);
       lib_char_to_i((WDX_BYTE *)&ewp->status,(WDX_INT *)&wp->status);
       lib_char_to_i((WDX_BYTE *)&ewp->validity,(WDX_INT *)&wp->validity);
*/
       lib_char_to_i((WDX_BYTE *)&ewp->index,&wp->index);
       lib_char_to_i((WDX_BYTE *)&ewp->status,&wp->status);
       lib_char_to_i((WDX_BYTE *)&ewp->validity,&wp->validity);
       }
}

/* void encode_resource()
 * Encode a resource structure into a encoded_resource structure
 *
 * Argument : pointer to resource structure
 *            pointer to encoded_resource structure
 */
void encode_resource(rp,erp)
struct resource *rp;
struct encoded_resource *erp;
{
   if ( rp && erp )
        {
        strncpy(erp->name,rp->name,RESOURCE_LEN);
//        lib_i_to_char((WDX_INT *)&rp->evttype,(WDX_BYTE *)&erp->evttype);
//        lib_i_to_char((WDX_INT *)&rp->value1,(WDX_BYTE *)&erp->value1);
//        lib_i_to_char((WDX_INT *)&rp->value2,(WDX_BYTE *)&erp->value2);
//        lib_i_to_char((WDX_INT *)&rp->value3,(WDX_BYTE *)&erp->value3);
        lib_i_to_char(&rp->evttype,(WDX_BYTE *)&erp->evttype);
        lib_i_to_char(&rp->value1,(WDX_BYTE *)&erp->value1);
        lib_i_to_char(&rp->value2,(WDX_BYTE *)&erp->value2);
        lib_i_to_char(&rp->value3,(WDX_BYTE *)&erp->value3);
        }
}

/* void decode_resource()
 * Decode a encoded_resource structure into a resource structure
 *
 * Argument : pointer to encoded_resource structure
 *            pointer to resource structure
 */
void decode_resource(erp,rp)
struct encoded_resource *erp;
struct resource *rp;
{
   if ( rp && erp )
        {
        strncpy(rp->name,erp->name,RESOURCE_LEN);
/*
		lib_char_to_i((WDX_BYTE *)&erp->evttype,(WDX_INT *)&rp->evttype);
        lib_char_to_i((WDX_BYTE *)&erp->value1,(WDX_INT *)&rp->value1);
        lib_char_to_i((WDX_BYTE *)&erp->value2,(WDX_INT *)&rp->value2);
        lib_char_to_i((WDX_BYTE *)&erp->value3,(WDX_INT *)&rp->value3);
*/
		lib_char_to_i((WDX_BYTE *)&erp->evttype,&rp->evttype);
        lib_char_to_i((WDX_BYTE *)&erp->value1,&rp->value1);
        lib_char_to_i((WDX_BYTE *)&erp->value2,&rp->value2);
        lib_char_to_i((WDX_BYTE *)&erp->value3,&rp->value3);
        }
}

/* void encode_heartbeat()
 * Encode a heartbeat structure into a encoded_heartbeat structure
 *
 * Argument : pt to heartbeat structure
 *            pt to encoded_heartbeat structure
 */
void encode_heartbeat(hp,ehp)
struct heartbeat *hp;
struct encoded_heartbeat *ehp;
{
int i;

    if ( hp && ehp )
         {
         encode_watchdog(&(hp->wat),&(ehp->wat));
         for(i=0;i<RESOURCE_COUNT;i++)
              encode_resource(&(hp->res[i]),&(ehp->res[i]));
         }
}

/* void decode_heartbeat()
 * Encode a encoded_heartbeat structure into a heartbeat structure
 *
 * Argument : pt to encoded_heartbeat structure
 *            pt to heartbeat structure
 */
void decode_heartbeat(ehp,hp)
struct encoded_heartbeat *ehp;
struct heartbeat *hp;
{
int i;

    if ( ehp && hp )
         {
         decode_watchdog(&(ehp->wat),&(hp->wat));
         for(i=0;i<RESOURCE_COUNT;i++)
              decode_resource(&(ehp->res[i]),&(hp->res[i]));
         }
}

/* void encode_command()
 * Encode a command structure into a encoded_command structure
 *
 * Argument : pt to command structure
 *            pt to encoded structure
 */
void encode_command(cp,ecp)
struct command *cp;
struct encoded_command *ecp;
{
   if ( cp && ecp )
        {
/*
        lib_i_to_char((WDX_INT *)&cp->index,(WDX_BYTE *)&ecp->index);
        lib_i_to_char((WDX_INT *)&cp->type,(WDX_BYTE *)&ecp->type);
        lib_i_to_char((WDX_INT *)&cp->value1,(WDX_BYTE *)&ecp->value1);
        lib_i_to_char((WDX_INT *)&cp->value2,(WDX_BYTE *)&ecp->value2);
        lib_i_to_char((WDX_INT *)&cp->value3,(WDX_BYTE *)&ecp->value3);
        lib_i_to_char((WDX_INT *)&cp->value4,(WDX_BYTE *)&ecp->value4);
*/
        lib_i_to_char(&cp->index,(WDX_BYTE *)&ecp->index);
        lib_i_to_char(&cp->type,(WDX_BYTE *)&ecp->type);
        lib_i_to_char(&cp->value1,(WDX_BYTE *)&ecp->value1);
        lib_i_to_char(&cp->value2,(WDX_BYTE *)&ecp->value2);
        lib_i_to_char(&cp->value3,(WDX_BYTE *)&ecp->value3);
        lib_i_to_char(&cp->value4,(WDX_BYTE *)&ecp->value4);
        strncpy(ecp->commanddata,cp->commanddata,COMMAND_LEN);
        }
}

/* void decode_command()
 * decode a encoded_command structure into a command structure
 *
 * Argument : pt to encoded structure
 *            pt to command structure
 */
void decode_command(ecp,cp)
struct encoded_command *ecp;
struct command *cp;
{
   if ( cp && ecp )
        {
/*
        lib_char_to_i((WDX_BYTE *)&ecp->index,(WDX_INT *)&cp->index);
        lib_char_to_i((WDX_BYTE *)&ecp->type,(WDX_INT *)&cp->type);
        lib_char_to_i((WDX_BYTE *)&ecp->value1,(WDX_INT *)&cp->value1);
        lib_char_to_i((WDX_BYTE *)&ecp->value2,(WDX_INT *)&cp->value2);
        lib_char_to_i((WDX_BYTE *)&ecp->value3,(WDX_INT *)&cp->value3);
        lib_char_to_i((WDX_BYTE *)&ecp->value4,(WDX_INT *)&cp->value4);
*/
        lib_char_to_i((WDX_BYTE *)&ecp->index,&cp->index);
        lib_char_to_i((WDX_BYTE *)&ecp->type,&cp->type);
        lib_char_to_i((WDX_BYTE *)&ecp->value1,&cp->value1);
        lib_char_to_i((WDX_BYTE *)&ecp->value2,&cp->value2);
        lib_char_to_i((WDX_BYTE *)&ecp->value3,&cp->value3);
        lib_char_to_i((WDX_BYTE *)&ecp->value4,&cp->value4);
        strncpy(cp->commanddata,ecp->commanddata,COMMAND_LEN);
        }
}

/* void encode_operation()
 * Encode a operation structure into a encoded_operation structure
 *
 * Argument : pt to operation structure
 *            pt to encoded structure
 */
void encode_operation(cp,ecp)
struct operation *cp;
struct encoded_operation *ecp;
{
   if ( cp && ecp )
        {
/*
        lib_i_to_char((WDX_INT *)&cp->index,(WDX_BYTE *)&ecp->index);
        lib_i_to_char((WDX_INT *)&cp->pid,(WDX_BYTE *)&ecp->pid);
        lib_i_to_char((WDX_INT *)&cp->type,(WDX_BYTE *)&ecp->type);
	lib_i_to_char((WDX_INT *)&cp->identifier,(WDX_BYTE *)&ecp->identifier);
        strncpy(ecp->operationdata,cp->operationdata,OPERATION_LEN);
        lib_i_to_char((WDX_INT *)&cp->operationargument,
                      (WDX_BYTE *)&ecp->operationargument);
*/
        lib_i_to_char(&cp->index,(WDX_BYTE *)&ecp->index);
        lib_i_to_char(&cp->pid,(WDX_BYTE *)&ecp->pid);
        lib_i_to_char(&cp->type,(WDX_BYTE *)&ecp->type);
	lib_i_to_char(&cp->identifier,(WDX_BYTE *)&ecp->identifier);
        strncpy(ecp->operationdata,cp->operationdata,OPERATION_LEN);
        lib_i_to_char(&cp->operationargument,
                      (WDX_BYTE *)&ecp->operationargument);
        }
}

/* void decode_operation()
 * decode a encoded_operation structure into a operation structure
 *
 * Argument : pt to encoded structure
 *            pt to operation structure
 */
void decode_operation(ecp,cp)
struct encoded_operation *ecp;
struct operation *cp;
{
   if ( cp && ecp )
        {
/*
        lib_char_to_i((WDX_BYTE *)&ecp->index,(WDX_INT *)&cp->index);
        lib_char_to_i((WDX_BYTE *)&ecp->pid,(WDX_INT *)&cp->pid);
        lib_char_to_i((WDX_BYTE *)&ecp->type,(WDX_INT *)&cp->type);
	lib_char_to_i((WDX_BYTE *)&ecp->identifier,
                      (WDX_INT *)&cp->identifier);
        strncpy(cp->operationdata,ecp->operationdata,OPERATION_LEN);
	lib_char_to_i((WDX_BYTE *)&ecp->operationargument,
                      (WDX_INT *)&cp->operationargument);
*/
        lib_char_to_i((WDX_BYTE *)&ecp->index,&cp->index);
        lib_char_to_i((WDX_BYTE *)&ecp->pid,&cp->pid);
        lib_char_to_i((WDX_BYTE *)&ecp->type,&cp->type);
	lib_char_to_i((WDX_BYTE *)&ecp->identifier,
                      &cp->identifier);
        strncpy(cp->operationdata,ecp->operationdata,OPERATION_LEN);
	lib_char_to_i((WDX_BYTE *)&ecp->operationargument,
                      &cp->operationargument);
        }
}

/* void update_group()
 * Using the new group data (from config files ), and the old snapshot
 * then update the group data base :
 *	for each member in new config file, search for information in the old
 *	snapshot. Update network information and resource information.
 *	Resource information is updated only if the member has same functional
 *	type.
 *	All the resources that have no owner at the end of the update group
 *	are marked with NOT_AN_INDEX as initiator value. After silence period
 *	eventscript will be trigerred for each of these resources.
 *
 * Argument : new group structure pt
 *	      old group structure pt
 *
 * Return   : None
 */
void update_group(n,o)
struct group *n;
struct group *o;
{
int _i;	/* index : new memberlist index */
int _j; /* index */
int _f; /* found : old memberlist index */
struct resource_def *_or; /* pt to old resource element */
struct resource_def *_nr; /* pt to new resource element */
struct member       *_om; /* pt to old member element */
struct member       *_nm; /* pt to new member element */

  /* set all resources to NOT_MARKED */
  for(_j=0;_j<RESOURCE_COUNT;_j++)
    o->resourcelist[_j].flag = NOT_MARKED;

  for(_i=0;_i<n->membercount;_i++)
  {
    /* search for member index in the old data base */
    _f = NOT_AN_INDEX;
    for ( _j=0;_j<o->membercount;_j++)
	if(strcmp(n->memberlist[_i].name,o->memberlist[_j].name) == 0 )
    {
	    _f = _j;
	    break;
    }
    /* if member registered in snapshot : update resource list */
    if ( _f != NOT_AN_INDEX )
    {
	/* copy network information */
	_om = &(o->memberlist[_f]);
	_nm = &(n->memberlist[_i]);
        /* restore status : brdstatus set to status for script trigerring */
	/* _nm->status = _om->status; */
	if ( _om->status != FAULTY )
	{
 	     _nm->status = IDLE;
             _nm->brdstatus = _om->status; /* restore status value */
        }
	else
        {
             _nm->status = FAULTY;
             _nm->brdstatus = FAULTY;
        }
	for(_j=0;_j<NETWORKS;_j++)
	{
	    if(_om->networks[_j].use == USED && _nm->networks[_j].use == USED )
	    {
	       _nm->networks[_j].validity = _om->networks[_j].validity;
	       _nm->networks[_j].status = _om->networks[_j].status;
	       _nm->networks[_j].mstatus = _om->networks[_j].mstatus;
	    }
	}

	/* copy resourcelist information */
	for(_j=0;_j<RESOURCE_COUNT;_j++)
	{
	    _or = &(o->resourcelist[_j]);
	    _nr = &(n->resourcelist[_j]);
	    if((_or->name[0] != '\0') && (_or->flag == NOT_MARKED))
	    {
		if((_or->initiator == _f) &&	/* same initiator and same */
		   (_om->type == _nm->type))	/* functional type */
		{
		   *_nr = *_or;		 /* copy data */
		   _nr->initiator = _i;  /* update initiator */
		   _or->flag = MARKED;   /* resource marked */
		} /* end copy resource */

	    } /* end new resource */
	} /* end resource loop */
    } /* end found */
  } /* end for */

  /* copy all the resources not marked : no more initiator */
  for(_j=0;_j<RESOURCE_COUNT;_j++)
  {
    _or = &(o->resourcelist[_j]);
    _nr = &(n->resourcelist[_j]);
    if((_or->name[0] != '\0') && (_or->flag == NOT_MARKED))
    {
	*_nr = *_or;		        /* copy data */
	_nr->initiator = NOT_AN_INDEX;  /* update initiator */
    } /* end copy resource */
  }

}

/* void process_communication_exit()
 * Exit handler
 */
void process_communication_exit()
{
  com_close_socket(socketdescriptor);
}
/* void process_info_exit()
 * Exit handler
 */
void process_info_exit(void)
{
   lib_delete_process_info(ggs.mission);
}

/* int get_the_index(hname)
 * Get the local index in group data table
 *
 * Argument : local name
 *
 * Return   : NOTANINDEX if not found / index
 */
int get_the_index(g,hname)
struct group *g;
char *hname;
{
int ret;
int node;

    ret = NOT_AN_INDEX;
    for(node=0; node < g->membercount; node++)
    {
	if (strcmp(hname,g->memberlist[node].name) == 0)
	{
	    ret = node;
	    break;
	}
    }

    return (ret);
}

/* void build_header()
 * Build a header message
 *
 * Argument : pt to header
 *            message type
 *            data size
 *            data number
 */
void build_header(p,t,s,n)
struct header *p;       /* header pointer */
enum msg_type   t;       /* data type */
int            s;       /* data size */
unsigned int   n;       /* data number */
{
  if(p)
     {
     p->wdxversion = wdxversion;
     p->configversion = configversion;
     p->missionkey = missionkey;
     p->memberindex = localindex;
     p->memberpid = getpid();
     p->datanumber = n;
     p->datatype = t;
     p->datasize = s;
     }
}

/* void build_watchdog()
 * Build watchdog message
 *
 * Argument : pt to watchdog
 *            pt to member
 */
void build_watchdog(p,m)
struct watchdog *p;
struct member *m;
{
   if(p && m)
      {
      p->index = localindex;
      p->validity = m->validity;
      p->status = m->status;
      }
}

/* int lib_build_generation()
 * Build a value for the generation number : current system time
 *
 * Argument : None
 *
 * Return   : generation number
 */
time_t lib_build_generation()
{
    return(spc_build_generation());
}

/* void lib_timer_reset()
 * Reset a wdx_timer structure
 *
 * Argument : pt to wdx_timer structure
 *
 * Return : None
 */
void lib_timer_reset(struct wdx_timer *p)
{
    if(p)
    {
	p->flag = FLAGRESET;
	p->event = EVENTRESET;
	p->value = END_VALIDITY;
    }
}

/* void lib_timer_set()
 * Set an wdx_timer structure to timeout passed as argument
 *
 * Argument : pt to wdx_timer structure
 *	      timeout
 *
 * Return : None
 */
void lib_timer_set(struct wdx_timer *p,int t)
{
    if(p)
    {
	p->flag = FLAGSET;
	p->event = EVENTRESET;
	p->value = t;
    }
}

/* void lib_timer_decrease()
 * Decrease an wdx_timer structure by val
 * If end of timeout is reached ( END_VALIDITY ) then the event flag is set.
 *
 * Argument : pt to wdx_timer structure
 *	      value to decrease
 *
 * Return : None
 */
void lib_timer_decrease(struct wdx_timer *p,int v)
{
    if(p)
    {
	if ( p->flag == FLAGSET)
	{
	    p->value -= v;
	    if( p->value <= END_VALIDITY )
	        p->event = EVENTSET;
	}
    }
}

/* int lib_timer_valid()
 * Test if an wdx timer is valid : flag set, event not set
 *
 * Argument : pt to wdx_timer structure
 *
 * Return : 1 if valid / 0 if invalid
 */
int lib_timer_valid(struct wdx_timer *p)
{
    if(p)
    {
	if ( p->flag == FLAGSET && p->event == EVENTRESET )
	     return(1);
    }
    return(0);
}

/* int lib_timer_alive()
 * Test if an wdx timer is alive : flag set
 *
 * Argument : pt to wdx_timer structure
 *
 * Return : 1 if alive / 0 if resetted
 */
int lib_timer_alive(struct wdx_timer *p)
{
    if(p)
    {
	if ( p->flag == FLAGSET )
	     return(1);
    }
    return(0);
}

/* void lib_timer_event()
 * return 1 if the event is set 0 otherwise
 *
 * Argument : pt to wdx_timer structure
 *
 * Return : 1 event set / 0 otherwise
 */
int lib_timer_event(struct wdx_timer *p)
{
    if(p)
    {
	if ( p->event == EVENTSET )
	    return(1);
    }
    return(0);
}

/* void lib_timer_reached()
 * return 1 if the timer has reached the value passed as argument
 * return 0 otherwise
 *
 * Argument : pt to wdx_timer structure
 *	      value
 *
 * Return : 1 value reached set / 0 otherwise
 */
int  lib_timer_reached(struct wdx_timer *p,int t)
{
    if(p)
    {
	if ( p->value <= t )
	    return(1);
    }
    return(0);
}

/* void lib_eventscript()
 * Trigger an eventscript
 *
 * Argument : pt to eventscript_data
 *
 * Return   : None
 */
void lib_eventscript(struct eventscript_data *p)
{
struct member_evt   *_m;
struct network_evt  *_n;
struct resource_evt *_r;
struct mismatch_evt *_i;
char     _s[STRING_LEN]; /* pt to script */
int                  _f; /* eventscript flag */
char                *_t; /* effective resource name */
char                *_o; /* effective owner name */
struct group        *_g;

   if(p)
   {
	_g = p->group;
	if ( _g->eventscript[0] == '\0' )
	     _f = 0;
	else _f = 1;
	if ( p->index != NOT_AN_OWNER )
	     _o = _g->memberlist[p->index].name;
        else _o = "unknown";
	switch(p->type)
	{
	case EVT_MEMBER :
	    _m = (struct member_evt *) &(p->data.m);
	    sprintf(messagestring,MEMBER_TRANSITION_FORMAT_MESSAGE,
		    _o,
		    member_status_to_s(_m->old),
		    member_status_to_s(_m->new));
	    sprintf(_s,MEMBER_TRANSITION_FORMAT_SCRIPT,
		    _g->eventscript,
		    _o,
		    member_status_to_s(_m->old),
		    member_status_to_s(_m->new));
	    if ( _f )
		 if ( _g->membermasks[_m->old][_m->new].value == NO )
		    _f = 0;
	    break;
	case EVT_NETWORK :
	    _n = (struct network_evt *) &(p->data.n);
	    sprintf(messagestring,NETWORK_TRANSITION_FORMAT_MESSAGE,
		    _o,
		    _n->net+NETWORK_OFFSET,
		    _g->memberlist[p->index].networks[_n->net].ifname,
		    network_status_to_s(_n->old),
		    network_status_to_s(_n->new));
	    sprintf(_s,NETWORK_TRANSITION_FORMAT_SCRIPT,
		    _g->eventscript,
		    _o,
		    _n->net+NETWORK_OFFSET,
		    _g->memberlist[p->index].networks[_n->net].ifname,
		    network_status_to_s(_n->old),
		    network_status_to_s(_n->new));
	    if ( _f )
		 if ( _g->networkmasks[_n->old][_n->new].value == NO )
		    _f = 0;
	    break;
	case EVT_RESOURCE :
	    _r = (struct resource_evt *) &(p->data.r);
	    if ( strcmp(_r->name,MASTER_TOKEN) == 0 )
	    {
		 _t = MASTER_TOKEN_STRING;
		 _f = 0;	/* no eventscript for mastertoken */
	    }
	    else _t = _r->name;
	    sprintf(messagestring,RESOURCE_TRANSITION_FORMAT_MESSAGE,
		    _t,
		    resource_status_to_s(_r->new),
		    _o);
	    sprintf(_s,RESOURCE_TRANSITION_FORMAT_SCRIPT,
		    _g->eventscript,
		    _t,
		    resource_status_to_s(_r->old),
		    resource_status_to_s(_r->new),
		    _o);
	    if ( _f )
		 if ( _g->resourcemasks[_r->old][_r->new].value == NO )
		    _f = 0;

	    break;
	case EVT_MISMATCH :
	    _i = (struct mismatch_evt *) &(p->data.i);
	    if ( strcmp(_i->name,MASTER_TOKEN) == 0 )
		 _t = MASTER_TOKEN_STRING;
	    else _t = _i->name;
	    sprintf(messagestring,MISMATCH_TRANSITION_FORMAT_MESSAGE,
		    _t,
		    _o);
	    sprintf(_s,MISMATCH_TRANSITION_FORMAT_SCRIPT,
		    _g->eventscript,
		    _t,
		    _o);
	    if ( _f )
		 if ( _g->mismatchmask.value == NO )
		    _f = 0;
	    break;
	}
	if ( p->script == YES && _f )
	    spc_trigger_script(_s,NULL);
        if ( p->display == YES )
	{
	    PINFOR_B
	    messagestring
	    PINFOR_E;
	}
   }
}

/* lib_is_numeric()
 * Check if a string represents a integer
 *
 * Argument : pt to string
 *
 * return   : 0 numeric / 1 non numeric
 *
 */
int lib_is_numeric(s)
char *s;
{
char *_c;
int _ret;

     _ret = 0;
     _c = s;
     while( *_c )
     {
           if ( *_c < '0' || *_c > '9' )
           {
                _ret = 1;
                break;
           }
           _c++;
     }
     return(_ret);
}
